"""Provides meteocons by Bas Milius for inkBoard. Not all icons are included in order to keep the file size down.
Not all icons are included. This integration is meant to be used by other integrations, and has no configuration to go along with it. 
To include it in a package, add meteocons: to your config.
"""

from typing import Literal
from pathlib import Path

import inkBoard

_LOGGER = inkBoard.getLogger(__name__)

def parse_weather_icon(condition: str, nighttime: bool = False, icon_type: Literal["filled","outline"] = "filled") -> Path:
    """Parses the appropriate meteocon icon for the provided condition

    Parameters
    ----------
    condition : str
        The weather condition to get the icon of
    nighttime : bool, optional
        Whether to look for for the nighttime version of the condition, by default False
        If the nighttime icon does not exist, returns the daytime version
    icon_type : Literal[&quot;filled&quot;,&quot;outline&quot;], optional
        The type of icon to return, by default "filled"

    Returns
    -------
    Path
        The path to the icon's file
    """    
    if condition in {"default", None}:
        icon_id = METEOCONS_WEATHER_ICONS["default"]

    elif nighttime:
        icon_id = METEOCONS_WEATHER_ICONS["night"].get(condition,METEOCONS_WEATHER_ICONS["day"].get(condition,"default"))
    else:
        icon_id = METEOCONS_WEATHER_ICONS["day"].get(condition,"default")

    if icon_id == "default":
        _LOGGER.warning(f"Could not find weather condition {condition} in the condition day keys, returning default value")
        icon_id = METEOCONS_WEATHER_ICONS["default"]

    icon_file = f"{icon_id}{IMAGE_FILE_TYPE}"
    if icon_type == "outline":
        return METEOCONS_PATH_OUTLINE / icon_file
    else:
        return METEOCONS_PATH_OUTLINE / icon_file
    
def parse_icon(icon: str, icon_type: Literal["filled","outline"] = "filled") -> Path:
    """Parses the path to a meteocon icon

    Parameters
    ----------
    icon : str
        The icon to parse
    icon_type : Literal[&quot;filled&quot;,&quot;outline&quot;], optional
        The type of icon to parse, by default "filled"

    Returns
    -------
    Path
        The path to the icon, or None if it was not found
    """

    if not "." in icon:
        icon_file = f"{icon}{IMAGE_FILE_TYPE}"
    else:
        icon_file = icon

    if icon_type == "outline":
        p = METEOCONS_PATH_OUTLINE / icon_file
    else:
        p = METEOCONS_PATH_OUTLINE / icon_file
    
    if not p.exists():
        _LOGGER.error(f"Meteocon icon {icon_file} does not exist")
        return
    
    return p

def setup(core, config):
    return True

METEOCONS_PATH_OUTLINE = Path(__file__).parent / "icons/meteocons/outline"
METEOCONS_PATH = Path(__file__).parent / "icons/meteocons/filled"

IMAGE_FILE_TYPE = ".png"
##File type of the included meteocons

METEOCONS_WEATHER_ICONS : dict = {"default": "cloudy",
        "day": {
            "clear-night": "clear-night",
            'cloudy':"overcast",
            "exceptional": "rainbow",
            'fog': "fog",
            'hail': "hail",
            'lightning': 'thunderstorms-extreme',
            "lightning-rainy": "thunderstorms-extreme-rain",
            "partlycloudy": "partly-cloudy-day",
            "pouring": "extreme-rain",
            'rainy': "overcast-drizzle",
            "snowy": "overcast-snow",
            "snowy-rainy": "overcast-sleet",
            "sunny": "clear-day",
            "windy": "umbrella-wind",
            "windy-variant": "umbrella-wind-alt",

            "hazy": "haze",
            "hurricane": "hurricane",
            "dust": "dust",
            "partly-lightning": "thunderstorms-day-overcast",
            "partly-rainy": "overcast-day-drizzle",
            "partly-snowy": "overcast-day-snow",
            "partly-snowy-rainy": "overcast-day-sleet",             
            "snowy-heavy": "extreme-snow",
            "tornado": "tornado"
            },
        "night": {
            "clear-night": "falling-stars",
            'cloudy':"overcast-night",
            "exceptional": "rainbow",
            'fog': "fog-night",
            'hail': "partly-cloudy-night-hail",
            'lightning': 'thunderstorms-night-extreme',
            "lightning-rainy": "thunderstorms-night-extreme-rain",
            "partlycloudy": "overcast-night",
            "pouring": "extreme-night-rain",
            'rainy': "overcast-night-drizzle",
            "snowy": "overcast-night-snow",
            "snowy-rainy": "overcast-night-sleet",
            "sunny": "falling-stars",

            "hazy": "overcast-night-haze",
            "dust": "dust-night",
            "partly-lightning": "thunderstorms-night-overcast",
            "partly-rainy": "partly-cloudy-night-drizzle",
            "partly-snowy": "partly-cloudy-night-snow",
            "partly-snowy-rainy": "partly-cloudy-night-sleet",             
            "snowy-heavy": "extreme-night-snow",
            }}
"Dict linking meteocon images to conditions. Suitable for both filled and outlined. Does not yet have the .png extension."

METEOCONS_FORECAST_ICONS : dict = {
                        "datetime" : None,
                        "cloud_coverage": "cloud-up",
                        "humidity": "humidity",
                        "apparent_temperature": "thermometer-sunny",
                        "dew_point": "thermometer-raindrop",
                        "precipitation": "raindrop-measurement",
                        "pressure": "barometer",
                        "temperature": "thermometer",
                        "templow": "thermometer-colder",
                        "wind_gust_speed": "wind-alert",
                        "wind_speed": "wind",
                        "precipitation_probability": "raindrop",
                        "uv_index": "uv-index",
                        "wind_bearing": "windsock"
                            }
"Meteocon icons for forecast entries."


"""
Handler to load inkBoard integrations present in the designer etc. folder.
"""

import importlib.util
from typing import Callable, TYPE_CHECKING, Literal, Any
from types import MappingProxyType
import logging
from pathlib import Path
import sys
import asyncio
import json

from inkBoard import loaders
from inkBoard.helpers import classproperty, reload_full_module

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from inkBoard import core as CORE

class IntegrationLoader(loaders.IntegrationLoader):
    "Provides bindings to load inkBoard integrations including those in the designer."

    @classproperty
    def integration_keys(cls) -> dict[str,str]:
        return cls._integration_keys.copy()

    @classmethod
    def get_integrations(cls, folders: dict[str,Path]):
        #folders: dict with base module name mapped to the folder path
        cls._reset()

        for base_module, folder in folders.items():
            if folder.exists():
                cls._read_out_folder(base_module, folder)

        return MappingProxyType(cls._installed_integrations)
    
    @classmethod
    def add_integration_config_key(cls, key: str, module_name: str):
        if key in cls._integration_keys:
            int_mod = cls._integration_keys[key]
            _LOGGER.info(f"{key} is already used for a the config of a different integration: {int_mod}")
        else:
            cls._integration_keys[key] = module_name    ##This should go the other way around.
        return

    @classmethod
    def _read_out_folder(cls, base_module: str, folder: Path):
        folders = folder.iterdir()
        for int_dir in filter(lambda p: (p.is_dir() and not p.name.startswith("_")), folders):
            if int_dir.name in cls._installed_integrations:
                _LOGGER.info(f"Integration {int_dir.name} has already been found in module {cls._integration_modules[int_dir.name]}. Will not import from {base_module}")
                continue

            manifest = int_dir / "manifest.json"
            if not manifest.exists():
                _LOGGER.error(f"Integration folder {int_dir} is missing the manifest.json file.")
                continue

            with open(manifest) as f:
                manifest = json.load(f)
                ##Support for requirements has not yet been implemented
            
            if c := manifest.get("config_entry",False):
                #Will require config_keys, similar to esphome, which can be left empty if needed.
                name = f"{base_module}.{int_dir.name}"
                                        

                ##These should not be checked by config key (i.e., save the key in the dict entry); key should be the folder name.
                cls.add_integration_config_key(c, name)
                cls._installed_integrations[int_dir.name] = int_dir
                cls._integration_modules[int_dir.name] = name
            else:
                _LOGGER.error(f"Integrations are required to have a config_entry key {int_dir.name} does not")
                continue

    @classmethod
    def import_integrations(cls, core: "CORE", progress_func=None, value_range=()):
        config = core.config
        import_set: set[tuple[str, Path]] = set()

        ##Import custom via appending to path.
        ##And I think just import via here, but make the set or something recognise the base package too.
        ##Or save strings to the modules instead of their folders.

        for config_entry in cls._integration_keys:
            if config_entry in config.configuration:
                import_set.add((config_entry, cls._integration_keys[config_entry]))
        
        if not import_set: 
            return

        if progress_func:
            progress_func(value_range[0] + 1, f"Importing {len(import_set)} integrations")
            step = int((value_range[1] - value_range[0] - 1)/len(import_set))
            progress = value_range[0] + 1

        for (config_key, name) in import_set:
            integration = name.split(".")[-1]
            if progress_func:
                progress = progress + step
                progress_func(progress, f"Importing integration {integration}")
            module = cls._import_integration(name)
            if not module:
                _LOGGER.warning(f"Unable to successfully import integration {integration} from {name}")
            else:
                cls._imported_modules[integration] = module

        cls._reload_imports = False
        return
        ##See code in hass core: https://github.com/home-assistant/core/blob/ab5ddb8edfb72d5f5915574f642eba93afc5abdc/homeassistant/loader.py#L1669

    @classmethod
    def _import_integration(cls, name):
            
        module = None
        integration = name.split(".")[-1]

        ##Simply put: I don't think it'll matter how you set this up, the __init__ will be imported regardless
        ##Messing with the package name and whatever will just cause issues with relative imports and stuff
        ##So: make different file, like integration.py or something, that provided the hooks for the integration_loader
        ##Or simply make writers use the init to provide the correct hooks *shrug*

        spec = importlib.util.find_spec(f"{name}.designer")
        if spec:
            if f"{name}.designer" in sys.modules and cls._reload_imports:
                try:
                        reload_full_module(name)
                        module = sys.modules.get(name,None)
                except ImportError as exce:
                    _LOGGER.error(f"Unable to reload integration {name}, will try to import it again", exc_info=exce)
                    sys.modules.pop(name)
                _LOGGER.debug(f"Integration {name} has designer module")
            elif f"{name}.designer" in sys.modules:
                return sys.modules.get(f"{name}.designer",None)
            
            try:
                return cls._import_designer_module(integration, spec)
            except ImportError:
                return
        else:
            return super()._import_integration(name)

    @classmethod
    def _import_designer_module(cls, integration, spec):
        #Imports and verifies an integration's designer module. 
        #Overwrites the base setup function if present in the designer module
        
        designer_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(designer_module) ##This step takes quite some time.
        base_module = designer_module.__name__.removesuffix(".designer")
        base_module = sys.modules.get(base_module)
        
        if hasattr(designer_module,"async_setup"):
            base_module.async_setup = designer_module.async_setup
        elif hasattr(designer_module, "setup"):
            ##Whilst it would be possible to make the designer setup etc. have priority even over base async_setup,
            ##I think it's better not to do so, since the designer module should try to exhibit roughly similar behaviour to the native integration
            base_module.setup = designer_module.setup
        
        if not hasattr(base_module,"async_setup") and not hasattr(base_module,"setup"):
            _LOGGER.warning(f"Integration {integration} is missing the required setup function")
            return
        
        if hasattr(designer_module,"async_start"):
            base_module.async_start = designer_module.async_start
        elif hasattr(designer_module,"start"):
            base_module.start = designer_module.start

        if hasattr(designer_module, "async_run"):
            base_module.async_run = designer_module.async_run

        return base_module

    @classmethod
    async def async_setup_integrations(cls, core: "CORE", progress_func=None, value_range=()) -> MappingProxyType[Literal["integration_entry"],Any]:
        integration_objects = {}
        
        config = core.config
        screen = core.screen

        if not cls._imported_modules:
            return MappingProxyType({})

        if progress_func:
            progress_func(value_range[0] + 1, f"Importing {len(cls._imported_modules)} integrations")
            step = int((value_range[1] - value_range[0] - 1)/len(cls._imported_modules))
            progress = value_range[0] + 1

        for integration, module in cls._imported_modules.items():
            module : "sys.ModuleType"
            setup_func = None
            if hasattr(module,"async_setup"):
                setup_func = module.async_setup
            elif hasattr(module,"setup"):
                setup_func = module.setup

            if not isinstance(setup_func,Callable):
                _LOGGER.error(f"{integration} does not have a valid setup function, not importing")
                raise TypeError
            
            _LOGGER.info(f"Setting up integration {integration}")
            if progress_func:
                progress = progress + step
                progress_func(progress, f"Setting up integration {integration}")
            try:
                if asyncio.iscoroutinefunction(setup_func):
                    res = await setup_func(core ,config)
                else:
                    res = setup_func(core, config)

                if res == None:
                    _LOGGER.error(f"Integration setup functions must return a result (at minimum a boolean `True`), or `False`. {integration} returned `None`")
                    continue

                if res == False:
                    _LOGGER.error(f"Something went wrong setting up {integration}")
                    continue
                
                cls._loaded_integrations.add(Path(module.__file__).parent)
                if res != True:
                    integration_objects[integration] = res

                _LOGGER.debug(f"{integration} succesfully set up")
            except (ModuleNotFoundError, ImportError) as exce:
                msg = f"Error importing integration {integration}: {exce}. The integration may not be able to run in an emulated environment."
                _LOGGER.warning(msg)
            except Exception as exce:
                msg = f"Error importing integration {integration}: {exce}."
                _LOGGER.error(msg, exc_info=True)

        return MappingProxyType(integration_objects)    
    
    @classmethod
    async def async_start_integrations(cls, core: "CORE"):
        """
        Calls the setup functions for relevant integrations. If a maximum time was set in the config, will continue setting up in the background while inkBoard starts printing.

        Parameters
        ----------
        core : CORE
            The inkBoard core module
        """   

        cls._done_setups = {}
        cls._pending_setups = {}

        config = core.config

        coro_list = set()
        for integration, module in cls._imported_modules.items():
            pkg = module.__package__
            setup_res = core.integration_objects.get(integration,None)

            if hasattr(module, "async_start"):
                if not asyncio.iscoroutinefunction(module.async_start):
                    _LOGGER.error(f"integration {integration}: async_start must be a coroutine")
                    continue
                t = asyncio.create_task(module.async_start(core, setup_res), name=pkg)

            elif hasattr(module, "start"):
                coro = asyncio.to_thread(module.start, core, setup_res)
                t = asyncio.create_task(coro,name=pkg)
            else:
                continue

            coro_list.add(t)
            cls._pending_setups[pkg] = t

        if not coro_list:
            return
        
        setup_time = config.inkBoard["integration_start_time"]

        if setup_time < 0:
            await cls._wait_for_start(coro_list)
        else:
            setup_task = asyncio.create_task(cls._wait_for_start(coro_list))
            try:
                await asyncio.wait_for(asyncio.shield(setup_task), setup_time)
            except asyncio.TimeoutError:
                _LOGGER.warning(f"Integration setup time of {setup_time} seconds elapsed but not all integrations are setup yet. Integrations {cls._pending_setups.keys()} will continue setup in the background.")

        return
    
    @classmethod
    async def _wait_for_start(cls, coro_list):
        pending = coro_list 
        while pending:
            try:
                done, pending = await asyncio.wait(pending,return_when=asyncio.FIRST_COMPLETED)
            except:
                for t in done:
                    t : asyncio.Future
                    if t.cancelled():
                        continue

                    if t.exception() != None:
                        _LOGGER.error(f"Integrations {t.get_name()} ran into an error while starting up: {t.exception()}")
            else:
                names = set()
                for t in done:
                    names.add(t.get_name())
                _LOGGER.info(f"Integrations {names} were started.")
            finally:
                cls._pending_setups.pop(t.get_name())
                cls._done_setups[t.get_name()] = t

    @classmethod
    async def run_integrations(cls, core: "CORE"):
        """
        Runs any long running tasks defined by integrations.
        Exceptions do not interfere with the print loop.

        Parameters
        ----------
        core : CORE
            The inkBoard core module
        """
        coro_list = set()

        async def await_setup(pkg, runner : asyncio.Task):
            setup_task = cls._pending_setups.get(pkg,None)
            _LOGGER.warning(f"Waiting to start {pkg} long running task until its setup has finished.")
            if setup_task != None:
                await setup_task
                await asyncio.sleep(0)
            await runner

        for integration, module in cls._imported_modules.items():

            pkg = module.__package__
            setup_res = core.integration_objects.get(integration,None)

            if hasattr(module, "async_run"):
                if not asyncio.iscoroutinefunction(module.async_run):
                    _LOGGER.error(f"integration {pkg}: async_run must be a coroutine")
                    continue

                if pkg in cls._pending_setups:
                    coro = await_setup(pkg, module.async_run(core, setup_res))
                else:
                    coro = module.async_run(core, setup_res)

                t = asyncio.create_task(coro, name=pkg)
                coro_list.add(t)

        if not coro_list:
            return

        pending = coro_list
        ##From what I could find, this is the only way to catch out exceptions when they happen (gather only returns when everything is done, or returns a single result)
        ##While keeping a reference to the original task (which doesn't seem to happen for as_completed I believe) -> that only also just throws an exception

        done = []
        while pending:
            try:
                done, pending = await asyncio.wait(pending,return_when=asyncio.FIRST_COMPLETED)
            except:
                for t in done:
                    t : asyncio.Future
                    if t.cancelled():
                        _LOGGER.warning(f"Integration {t.get_name()} was cancelled while running.")

                    if t.exception() != None:
                        _LOGGER.error(f"Integration {t.get_name()} ran into an error while running: {t.exception()}")





"Some utilities to use with the designer"

import logging
from typing import *
import asyncio
import inspect
from threading import ExceptHookArgs

from PIL import ImageTk

if TYPE_CHECKING:
    from .tkinter.windows import DesignerWindow

_LOGGER = logging.getLogger(__package__)

main_loop: asyncio.BaseEventLoop = None
window: "DesignerWindow"

iidType = TypeVar("iid", bound=str)

def threading_except_hook(*excepthooksargs):
    if not excepthooksargs: return
    args: ExceptHookArgs = excepthooksargs[0]
    if args.thread == window._inkBoard_thread:
        if window._inkBoard_lock.locked():
            window._inkBoard_lock.release()
    if args.exc_type not in {SystemExit, asyncio.CancelledError, RuntimeError}:
        if args.exc_value: raise args.exc_value
        else: raise args.exc_type()
    return


def call_in_main_thread(func : Callable, args, kwargs):
    
    if asyncio._get_running_loop() != main_loop != None:
        fut = asyncio.run_coroutine_threadsafe( #@IgnoreExceptions
            __call_in_main_thread(func,*args,**kwargs),
            main_loop)
        return fut.result()
    else:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            _LOGGER.error(f"Calling {func} with args {args} and kwargs {kwargs} caused an exception: {e}")
            raise e
    
async def __call_in_main_thread(func, *args, **kwargs):
    return func(*args, **kwargs)

class DummyTask:
    """
    Provides a dummy to mimic an asyncio task object when needed to make one before starting the event loop. For use in logic statements
    """
    
    def done(self) -> bool:
        """Returns True to mimic the task being done"""
        return True
    
    def cancelled(self) -> bool:
        """Returns False since the dummy task cannot be cancelled"""
        return False
    
    def cancel(self) -> None:
        """Does nothing but may be useful for logic purposes"""
        return
    
    def result(self) -> None:
        """Returns nothing since there is no result"""
        return

class ThreadSafeWidgetMeta(type):
    """Metaclass to use with subclassed tkwidgets. Wraps all object methods (not class or staticmethods!) into a threadsafe call for inkBoard designer.
    
    Generally not needed, since the designer used a thread-safe third party library for tkinter. However, for photoimages especially, this implementation functioned a lot faster.
    Be mindful using this as blocking call will block the event loop.
    """    

    def __new__(cls, name, bases, attrs):
        obj = super().__new__(cls, name, bases, attrs)
        for attr_name, attr_value in inspect.getmembers(obj, predicate=callable):
            if (attr_name.startswith("__") and attr_name.endswith("__") or
                attr_name == "cget"): ##Idk but cget causes issues with the TreeFrame at least and does not seem to interfere with the loop?
                continue

            if isinstance(attr_value,(classmethod,staticmethod)):
                continue

            setattr(obj,attr_name, cls.intercept_method_call(attr_value))
        return obj
    
    @staticmethod
    def intercept_method_call(method):
        def wrapper(*args, **kwargs):
            result = call_in_main_thread(method, args, kwargs)
            _LOGGER.verbose(f"Intercepted call to {method}")
            return result
        _LOGGER.verbose(f"Wrapped method {method}")
        return wrapper
    
class ThreadSafePhotoImage(ImageTk.PhotoImage, metaclass=ThreadSafeWidgetMeta):
    """Threadsafe implementation for calling photoimage, by passing the __init__ onto the main thread.
    
    Importing this module automagically changes the ImageTk.PhotoImage class to this one, to have it work correctly this module needs to be imported before imageTk is imported anywhere.
    Whilst tkthread also takes care of this, it appears to be a lot slower.
    """

    def __init__(self, image = None, size = None, **kw):
        call_in_main_thread(super().__init__,
                            (image, size), kw)



##Make the main window in here

from . import windows, widgets, functions

window = windows.DesignerWindow()
"The main window instance"

windows.window = window

functions.window = window
functions.windows = windows
functions.widgets = widgets

widgets.window = window


"""
This module (will) serve as an extension to the actual inkBoard module.
The designer is meant to help design and create inkBoard packages, for example provide an interface to "debug" your dashboard, but also hold the functionality to create e.g. standalone executables.
This way, the inkBoard module itself should be able to stay at a relatively small size, which should hopefully make it more portable.
"""
__version__ = "0.2.0"

from typing import TYPE_CHECKING
import logging
import asyncio
import threading
import sys

from . import const

if TYPE_CHECKING:
    import argparse

_LOGGER = logging.getLogger(__package__)

def check_threads() -> None:
    """Check if there are any lingering threads.

    From the HA source code
    """
    try:
        nthreads = sum(
            thread.is_alive() and not thread.daemon for thread in threading.enumerate()
        )
        if nthreads > 1:
            sys.stderr.write(f"Found {nthreads} non-daemonic threads.\n")

    except AssertionError:
        sys.stderr.write("Failed to count non-daemonic threads.\n")


def run_designer(args):
    """
    Runs inkBoard designer and builds the window

    Parameters
    ----------
    args : _type_
        Command line arguments
    """    

    ##Make base config and dashboard config into singleton classes
    ##At least take out the classs structure as it allows for more flexibility
    ##Also, don't forget to call the reload function when pressing the quit button

    ##Also don't forget, when running a new config to i.e. rebuild the icons and stuff

    from .runners import async_run_designer, async_stop_designer
    from inkBoard import logging as ib_logging

    ib_logging.init_logging()

    try:
        asyncio.run(async_run_designer(args))
    except KeyboardInterrupt:
        try:
            asyncio.run(async_stop_designer())
        except:
            pass

    check_threads()
    return 0


def _add_parser(parser: "argparse._SubParsersAction", name: str):
    designer_parser: "argparse.ArgumentParser" = parser.add_parser(name, 
                    description="Run inkBoard in designer mode, meant for desktop environments. Allows for designing dashboards without being limited by the platform it will run on, but with emulation capabilities to make performance as close as possible to what it will be on device. Requires inkBoarddesigner to be installed",
                    help="Runs inkBoard in designer mode", add_help=True)
    
    designer_parser.add_argument(const.ARGUMENT_CONFIG, nargs="?",
                                help="The YAML file used for the dashboard", default=None)

    return



import asyncio
import logging
from typing import *
from types import MappingProxyType
import tkinter as tk
from pathlib import Path
from dataclasses import asdict
import threading

import tkthread

from PIL import Image

import ttkbootstrap as ttk
from ttkbootstrap.tooltip import ToolTip

from PythonScreenStackManager.devices import FEATURES

from .widgets import Treeview, PSSMCanvas, BatteryFrame, BacklightFrame
from . import functions as tk_functions

from .. import const, util
from ..settings import EM_SETTINGS
from ..const import REFRESH_RATE

if TYPE_CHECKING:
    from PythonScreenStackManager.elements import Element
    from inkBoarddesigner.emulator.device import Device
    from inkBoard import core

_LOGGER = logging.getLogger(__name__)

window: "DesignerWindow"
"The main window instance"

p = Path()


class DesignerWindow(ttk.Window):
    """Specific window class meant to interface with inkBoard Designer
    
    A subclass of a ttk.Window. Sets up some base variables and functions for default behaviour.
    All calls to tkinter are automatically forwarded to the correct thread.
    """

    def __init__(self, title="ttkbootstrap", themename="litera",  size=None, position=None, minsize=None, maxsize=None, resizable=None, hdpi=True, scaling=None, transient=None, overrideredirect=False, alpha=1):
        super().__init__(title, themename, None, size, position, minsize, maxsize, resizable, hdpi, scaling, transient, overrideredirect, alpha)

        util.window = self

        window_icon = Path(__file__).parent.parent / "files" / "inkboard_designer_icon.ico"
        self.wm_iconbitmap(window_icon, window_icon)

        self._current_config_file = None

        if EM_SETTINGS.getboolean(const.DARKMODE_VAR_NAME):
            themes = (const.THEME_LIGHT,const.THEME_DARK)
        else:
            themes = (const.THEME_DARK, const.THEME_LIGHT)
        
        tk_functions.configure_themes(themes)
        self.style.theme_use(themes[-1])

        self.__state = ""
        self.__loaded_event = asyncio.Event()

        self._mainLoop: asyncio.BaseEventLoop = None
        self._inkBoard_thread: threading.Thread = None
        self._inkBoard_clean = True 
        #Indicates the last inkBoard thread was cleaned up (i.e. fully reloaded)

        self._inkBoard_lock = threading.Lock()
        "Lock to ensure only one inkBoard lock runs at a time"

        self._darkmode_variable = tk.BooleanVar(self, value=EM_SETTINGS.getboolean(const.DARKMODE_VAR_NAME), name=const.DARKMODE_VAR_NAME)
        self._saveasvariable = tk.BooleanVar(self, value=EM_SETTINGS.getboolean(const.SAVEAS_VAR_NAME), name=const.SAVEAS_VAR_NAME)
        self._hightlight_variable = tk.BooleanVar(self, value=EM_SETTINGS.getboolean(const.HIGHLIGHT_VAR_NAME), name=const.HIGHLIGHT_VAR_NAME)

        self._tree_list_variable = ttk.StringVar(self, name=const.LIST_VAR_NAME, 
                                                value=const.NO_TREE_OPTION)

        self._window_update_task : asyncio.Task = None
        self._resizeTask: asyncio.Task = util.DummyTask()
        
        self._screenCanvas = PSSMCanvas(self, name=const.CANVAS_NAME, cursor="target")
        self._screenCanvas.pack(fill=tk.BOTH, side=tk.LEFT)

        self._background_Tk_idx = None

        self._resize_bind = None
        self._width = self.winfo_width()
        self._height = self.winfo_height()
        self.bind("<Configure>", self._configure, add="+")

        self._reloading = True

        self.interface_buttons : list[ttk.Button] = []

        vars = (self._darkmode_variable, self._saveasvariable, self._hightlight_variable, self._tree_list_variable)

        for var in vars:
            var.trace_add("write", self.trace_variable)

        self._buttonLabel: ttk.Label

        self._progressframe = ttk.Frame(self.screenCanvas, bootstyle=ttk.SECONDARY)
        
        self._progressTitle = ttk.Label(self._progressframe, font=const.DEFAULT_LABEL_FONT, text="Loading emulator", 
                                    style=f"{ttk.PRIMARY}.{const.LOADBAR_LABEL_STYLE}")
        self._progressBar = ttk.Progressbar(self._progressframe, bootstyle=(ttk.STRIPED, ttk.PRIMARY))
        self._progressText = ttk.Label(self._progressframe, style=const.LOADBAR_LABEL_STYLE)

        self._progressTitle.pack(side=tk.TOP)
        self._progressBar.pack(side=tk.TOP, fill=tk.X, padx=(10,10))
        self._progressText.pack(side=tk.BOTTOM)

    #region
    @property
    def hightlight_variable(self) -> tk.BooleanVar:
        "The variable tracking the highlight setting"
        return self._hightlight_variable
    
    @property
    def saveas_variable(self) -> tk.BooleanVar:
        "Variable tracking if the screenshots should show a dialog box before saving"
        return self._saveasvariable
    
    @property
    def darkmode_variable(self) -> tk.BooleanVar:
        "Variable tracking if darkmode is on"
        return self._darkmode_variable
    
    @property
    def tree_list_variable(self) -> ttk.StringVar:
        "Variable tracking which tree is to be displayed"
        return self._tree_list_variable
    
    @property
    def screenCanvas(self) -> PSSMCanvas:
        "Canvas that the PSSM screen prints onto"
        return self.children[const.CANVAS_NAME]

    @property
    def buttonFrame(self) -> ttk.LabelFrame:
        "Frame holding the main buttons"
        return self._buttonFrame
    
    @property
    def settingsFrame(self) -> ttk.LabelFrame:
        "Frame holding the settings widgets"
        return self._settingsFrame

    @property
    def treeFrame(self) -> "TreeFrame":
        "The frame holding the treeviews"
        return self._treeFrame
    
    @property
    def configLabel(self) -> ttk.Label:
        "The button to indicate the loaded config"
        return self._configLabel
        
    @property
    def state(self) -> Literal[ttk.ACTIVE, ttk.DISABLED]:
        "Current state of the interface. Disabled means certain widgets are disables."
        return self.__state

    @property
    def loaded(self) -> bool:
        "Whether an inkBoard config has been successfully loaded and is currently running"
        return self.__loaded_event.is_set()
    #endregion


    async def run_update_loop(self):
        "Starts the update loop. Relatively akin to running to running window.Mainloop, but async"
        self._mainLoop = asyncio.get_running_loop()
        util.main_loop = asyncio.get_running_loop()
        self._window_update_task = asyncio.create_task(self._update_loop(), name="designer-window-loop")
        try:
            await self._window_update_task
        except asyncio.CancelledError:
            pass
        return

    async def _update_loop(self):
        sleep_time = 1/REFRESH_RATE
        try:
            while True:
                    self.update()
                    await asyncio.sleep(sleep_time)
        except asyncio.CancelledError:
            self.destroy()
            return

    def stop_update_loop(self):
        "Stops the window update loop, and destroys it if running. **Does not handle actually quitting the programme!**"
        if self._window_update_task:
            self._window_update_task.cancel()

    def call_in_main_thread(self, func : Callable, *args, **kwargs):
        """Helper function that allows calls function `func` with arguments `args` and keyword arguments `kwargs` in the main loop.

        Needed since tkinter doesn't like having window settings changed in different threads.
        Does not work with Coroutines.
        Keep in mind a blocking function blocks the main thread, which means the window becomes unresponsive.
        """

        return util.call_in_main_thread(func, args, kwargs)

    def _configure(self, event : tk.Event):
        if event.widget != self:
            return
        
        if event.width != self._width or event.height != self._height:
            if abs(1 - (event.width/self._width)) > 0.05 or abs(1 - (event.height/self._height)) > 0.05:
                ##Larger increase than this: assume toggle fullscreen, so update right asap
                if self._resizeTask.done():
                    resize_event = None
                else:
                    return
            else:
                if not self._resizeTask.done():
                    return
                resize_event = asyncio.Event(loop=self._mainLoop)
                self._resize_bind = self.bind("<ButtonRelease-1>", lambda event: resize_event.set(), add="+") ##Won't add this to bind as it is removed upon releasing the mouse button

            if not self._mainLoop:
                self.screenCanvas._build_canvas_background()
                return
            
            self._resizeTask = self._mainLoop.create_task(self._resize(resize_event))
        return

    async def _resize(self, event : Union[tk.Event,asyncio.Event]):
        if event == None:
            pass
        elif isinstance(event,asyncio.Event):
            #The timeout may not really be needed since it seems resizing blocks the event loop anyways
            #But just to be sure it's useful to have imo
            _, pending = await asyncio.wait([event.wait()],timeout=0.25)
            if pending:
                event.set()
            self.unbind("<ButtonRelease-1>")
        else:
            return
        
        self._width = self.winfo_width()
        self._height = self.winfo_height()

        self._screenCanvas["width"] = self._width - const.INTERFACE_WIDTH
        self._screenCanvas["height"] = self._height
        self.screenCanvas.update()  ##Gotta update the canvas first to ensure it gathers the right size.
        self.screenCanvas._build_canvas_background()

        self.unbind("<ButtonRelease-1>", self._resize_bind)
        self._resize_bind = None
        return

    def setup_canvas(self, size):
        "Performs the inital canvas setup by configuring the width and creating the background image"
        return self.__setup_canvas(self, size)

    @tkthread.called_on_main
    def __setup_canvas(self, size):
        "Performs the inital canvas setup by configuring the width and creating the background image"

        self._screenCanvas["width"] = size[0]
        self._screenCanvas["height"] = size[1]

        self.screenCanvas._build_canvas_background()
        return self._screenCanvas
        
    
    def clear_canvas(self, *args):
        "Clears the screen canvas from everything except the background"
        tkthread.call_nosync(self.screenCanvas._clear)


    def set_progress_bar(self, value, text=None, title=None):
        self._set_progress_bar(self, value, text, title)
        return
    
    @tkthread.called_on_main
    def _set_progress_bar(self, value, text, title=None):
        
        if value == ttk.DANGER:
            self._progressBar.configure(bootstyle=(ttk.STRIPED, ttk.DANGER))
            self._progressTitle.configure(style=f"{ttk.DANGER}.{const.LOADBAR_LABEL_STYLE}")
            if text:
                self._progressText["text"] = text
            return
        
        if value < 0:
            self._progressframe.place_forget()
            self._progressText["text"] = ""
            self._progressTitle["text"] = ""
            self._progressBar["value"] = 0
            self._progressBar.configure(bootstyle=(ttk.STRIPED, ttk.PRIMARY))
            self._progressTitle.configure(style=f"{ttk.PRIMARY}.{const.LOADBAR_LABEL_STYLE}")
            return
        
        if value >= 100:
            async def hide_progress_bar():
                await asyncio.sleep(0.5)
                self.set_progress_bar(-1,"")
            self._mainLoop.create_task(hide_progress_bar())

        self._progressBar["value"] = value
        if text:
            self._progressText["text"] = text
        if title:
            self._progressTitle["text"] = title

        if not self._progressframe.winfo_viewable():
            self._progressframe.place(relx=0.25, rely=0.5,
                                    relwidth=0.5, height=100)
            self._progressframe.update()
            self._progressText.configure(wraplength=self._progressframe.winfo_width())
    
    def set_inkboard_state(self, state: Literal[ttk.ACTIVE, ttk.DISABLED, ttk.DANGER, "ERROR"]):
        "Sets the state of the inkBoard connected widgets"

        if state and state.upper() == "ERROR": 
            state = ttk.DANGER

        if state == self.__state:
            return
        
        self.__state = state
        if state in {ttk.DISABLED, None}:
            self._set_disabled_state(self)
            if state == None:
                self._set_config_label(self, state)
        elif state == ttk.ACTIVE:
            self._set_active_state(self)
        elif state == ttk.DANGER:
            self._set_button_state(self, ttk.DANGER)
            self._set_progress_bar(self, state, text=None)


    @tkthread.called_on_main
    def _set_disabled_state(self):
        #Set when no config is currently (succesfully) loaded
        self.__loaded_event.clear()
        self._set_button_state(self, ttk.DISABLED)
        self.clear_canvas()
        self.treeFrame._clean_up()
        self.set_progress_bar(-1)

    @tkthread.called_on_main
    def _set_active_state(self):
        self.__loaded_event.clear()
        self._set_button_state(self, ttk.ACTIVE)
        self.treeFrame._setup()
        return

    @tkthread.called_on_main
    def _set_button_state(self, state: Literal[ttk.ACTIVE, ttk.DISABLED, ttk.DANGER]):
        
        if state == ttk.DANGER:
            set_buttons = filter(lambda x: x.winfo_name() in const.ERROR_ACTIVE_BUTTONS, self.interface_buttons)
            button_state = ttk.ACTIVE
            self._set_config_label(self, state)
        else:
            set_buttons = self.interface_buttons
            button_state = state

        for button in set_buttons:
            button.configure(state=button_state)    #@IgnoreException

        self.treeFrame.list_menu.configure(state=button_state)
        self._set_config_label(self, state)

    @tkthread.called_on_main
    def _set_config_label(self, state):
        
        ##Use this one too to set the label to error style if loading the config errors.
        if state == ttk.ACTIVE:
            frame_state = ttk.PRIMARY
        elif state == ttk.DISABLED:
            frame_state = ttk.SECONDARY
        elif state == ttk.DANGER:
            frame_state = state
        elif state == None:
            self._current_config_file = None
            self.configLabel.pack_forget()
            self._buttonLabel.pack(side=tk.LEFT, fill=tk.X, expand=1)
            return

        self._buttonLabel.pack_forget()
        self.configLabel.pack(side=tk.LEFT, fill=tk.X, expand=1)
        self.configLabel.configure(style=f"{frame_state}.TLabel")

    def trace_variable(self, var_name, *args):
        newVal = window.globalgetvar(var_name) 
        EM_SETTINGS[var_name] = str(newVal)
        return

class TreeFrame(ttk.Frame):

    tree: ttk.Treeview
    "The current treeview widget"

    _window: DesignerWindow

    def __init__(self, master = None, **kwargs):
        super().__init__(master, **kwargs)

        self._base_trees = {"none": None}
        
        self.__current_tree_option = "none"
        self.__tree = None


        self.__registered_trees = {}

        self.bind("<Leave>", self._leave)

        scrollbar = ttk.Scrollbar(self, orient="vertical",
                                style=const.SCROLLBAR_STYLE)
        self.__scrollbar = scrollbar

        self.__base_options = (const.NO_TREE_OPTION, const.ELEMENT_TREE_OPTION)
        self.list_menu["values"] = self.__base_options
        self.list_menu.bind('<<ComboboxSelected>>', self._select_tree)
        
        val = EM_SETTINGS.get(const.LIST_VAR_NAME)
        self._setup_value = val

        self.list_menu.set(const.NO_TREE_OPTION)
        Treeview._treeframe = self
        return

    #region
    @property
    def tree(self) -> Optional[Treeview]:
        return self.__tree

    @property
    def current_option(self) -> str:
        "The currently selected option in the menu list"
        val_idx = self.list_menu.current()
        return self.list_menu["values"][val_idx]

    @property
    def current_tree_option(self) -> str:
        return self.__current_tree_option

    @property
    def list_menu(self) -> ttk.Combobox:
        "Menu holding the tree options"
        return window._list_menu

    @property
    def scrollbar(self) -> ttk.Scrollbar:
        "Scrollbar to scroll through trees"
        return self.__scrollbar

    @property
    def registered_trees(self) -> dict[str,Treeview]:
        return self._base_trees | self.__registered_trees

    @property
    def entity_tree(self) -> Treeview:
        return self.children[const.ENTITY_TREE_NAME]
    
    @property
    def element_tree(self) -> Treeview:
        return self._element_tree
    #endregion

    def _select_tree(self, event: tk.Event):
        val = self.current_option
        self.show_tree(val)
        return

    def show_tree(self, option: str):
        option = option.lower()
        if option not in self.registered_trees:
            _LOGGER.warning(f"No tree registered under option {option}")
            return
        
        new_tree = self.registered_trees[option]

        if self.tree:
            self.tree.pack_forget()
            self.tree.tooltip.hide_tip()
        else:
            self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        self.__current_tree_option = option
        self.__tree = new_tree

        if not new_tree:
            self.scrollbar.pack_forget()
            return
        
        new_tree.pack(in_ = self ,anchor=tk.S, fill=tk.BOTH, expand=1)
        self.scrollbar.configure(command=new_tree.yview)
        new_tree.configure(yscrollcommand=self.scrollbar.set)
        
        return

    def register_tree(self, tree_option: str, widget: Treeview):

        tree_option = tree_option.lower()
        assert isinstance(widget, Treeview), "Tree widget must be an inkBoard Treeview"
        assert tree_option not in self.registered_trees, f"A treewidget under option {tree_option} is already registered"

        self.__registered_trees[tree_option.lower()] = widget
        ops = list(self.list_menu["values"])
        ops.append(tree_option.title())
        self.list_menu["values"] = ops

    def get_current_tree(self) -> Optional[Treeview]:
        return self.tree
    
    def get_tree(self, tree_option: str) -> Treeview:
        option = tree_option.lower()
        if option == const.ELEMENT_TREE_OPTION.lower():
            return self._element_tree
        
        if option not in self.registered_trees:
            _LOGGER.warning(f"No tree registered under option {option}")
            return
        return self.registered_trees[option]

    def _leave(self, event):
        
        for tree in self.registered_trees.values():
            if tree == None: continue
            tree.tooltip.hide_tip()

    def _clean_up(self):
        _LOGGER.debug("Cleaning up treeframe")
        if self.tree:
            self.tree.disable()

        if hasattr(self, "_old_register"):
            self._setup_value = self.list_menu.get()

        self.list_menu.configure(state=ttk.DISABLED)

        self._old_register = self.__registered_trees.copy()
        self.__registered_trees = {}    ##Emptying this does remove the tree entirely. Maybe make a copy for an old_register?

        self.list_menu["values"] = self.__base_options
        self._element_tree: Treeview
        self._element_tree.delete(*self._element_tree.get_children())

    def _setup(self):
        
        self.list_menu.configure(state=ttk.READONLY)
        list_val = self._setup_value
        self._old_register = {}
        if list_val.lower() in self.registered_trees:
            tree_option = list_val
        else:
            tree_option = const.NO_TREE_OPTION

        self.list_menu.set(tree_option)
        self.show_tree(tree_option)
        self._old_register = {}
        return

class _AdditionalWindow(ttk.Toplevel):
    ##Used for additional windows of which only 1 instance is supposed to be open.

    _instance: "ElementWindow" = None

    __instances = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance:
            try:
                cls._instance.destroy()
            except:
                pass
            cls._instance = None
        cls._instance = super().__new__(cls)
        _AdditionalWindow.__instances[cls] = cls._instance
        return cls._instance

    @classmethod
    def close_all(cls):
        for instance_cls, instance in cls.__instances.items():
            if instance != None:
                try:
                    instance.destroy()
                except:
                    pass
            cls.__instances[instance_cls] = None
            instance_cls._instance = None
        return        


class ElementWindow(_AdditionalWindow):
    
    def __init__(self, element: "Element", **kwargs):

        self._element = element
        title = f"{element.__class__.__name__}: {element.id}"

        super().__init__(title, master=window,  **kwargs)

        elt_text = self.create_element_attribute_list()

        ##will probably turn this into a grid instead of using pack.

        text_widget = ttk.ScrolledText(self)
        text_widget.insert("0.0", elt_text)
        text_widget.configure(state=ttk.DISABLED)
        text_widget.grid(row=0, column=0, columnspan=2)
        img_button = ttk.Button(self,text="Show Image", command=self.show_element_image)
        img_button.grid(row=1, column=0)
        ToolTip(img_button,const.SHOW_IMAGE_TIP, const.TOOLTIP_STYLE)

        save_button = ttk.Button(self,text="Save Image", command=self.save_element_image)
        save_button.grid(row=1, column=1)
        ToolTip(save_button,const.SAVE_IMAGE_TIP, const.TOOLTIP_STYLE)

        self._img_button = img_button

        return

    def create_element_attribute_list(self) -> str:
        
        elt = self._element
        text = ""
        for attr in dir(elt):
            if attr[0] == "_":
                continue

            if attr == "id":
                continue

            if not hasattr(elt,attr):
                continue

            val = getattr(elt,attr,False)
            if callable(getattr(elt,attr)):
                continue
            
            if isinstance(val,(bool,str,int,float,dict,MappingProxyType, set, type(None))):
                pass
            elif isinstance(val,(list,tuple)):
                t = type(val).__name__
                val = str(val)
                if len(val) > 50:
                    val = f"{t}: [...]"
            else:
                val = val.__class__.__name__
                v = val

            text = text + "\n" + "   " + f"{attr}: {val}"
        return text

    def show_element_image(self, *args):
        
        if isinstance(self._element.imgData,Image.Image):
            self._element.imgData.show()

    def save_element_image(self, *args):
        
        if isinstance(self._element.imgData,Image.Image):
            img = self._element.imgData.copy()
            tk_functions.save_image(img, f"{self._element.id}_")

class DeviceWindow(_AdditionalWindow):
    
    def __init__(self, device: "Device", **kwargs):

        self._device = device
        if device == None:
            title = "Emulated Device"
            device_text = "No device being emulated"
        else:
            title = device._model
            device_text = self.create_device_text(device)

        super().__init__(title, master=window,  **kwargs)

        feature_pad_x = (10,10)
        h = device_text.count("\n") + 2

        self.grid_rowconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1, pad=10)

        if device != None:
            feature_col = 0
            feature_row = 1
            if device.has_feature(FEATURES.FEATURE_BATTERY):
                BatteryFrame(self, device).grid(sticky=ttk.NSEW, padx=feature_pad_x,
                                                            row=feature_row, column=feature_col)
                feature_col += 1

            if device.has_feature(FEATURES.FEATURE_BACKLIGHT): 
                if feature_col != 0:
                    self.add_seperator(feature_row, feature_col)
                    feature_col += 1

                BacklightFrame(self, device).grid(sticky=ttk.NSEW, padx=feature_pad_x,
                                                            row=feature_row, column=feature_col)
                feature_col += 1

        textw = ttk.Text(self, height = h)
        textw.insert("0.0", device_text)
        textw.configure(state=ttk.DISABLED)
        textw.grid(sticky=ttk.NSEW, pady=(5,10),
                row=0, column=0, columnspan=feature_col)

    def create_device_text(self, device: "Device"):
        text = f"{device.model}"
        
        attr_list = {"Platform": "_emulated_platform", "Model": "model", "Screen type": "screenType", 
                    "Width": "width", "Height": "height"}

        for t, attr in attr_list.items():
            val = getattr(device,attr,None)
            text = f"{text}\n    {t}: {str(val)}"
        
        feature_text = ""
        for feature, val in asdict(device._features).items():
            if val: 
                if feature_text:
                    feature_text = f"{feature_text}, {feature}"
                else:
                    feature_text = f"{feature}"
        
        text = f"{text}\n    Features: {feature_text}"

        return text
    
    def add_seperator(self, row, col):
        sep = ttk.Separator(self,orient="vertical", bootstyle=const.FEATURE_FRAME_STYLE)
        sep.grid(row=row, column=col, sticky=ttk.NSEW, pady=(0,10))

class ConfigWindow(_AdditionalWindow):

    def __init__(self, core: "core", **kwargs):
        ##Need to determine what to show in this window
        ##At least: config name; platform; loaded integrations
        ##Also: allow for installing integration/platform requirements from here
        ##Likely: also the ui to run examples

        self._core = core

        super().__init__("Configuration", **kwargs)
        label = ttk.Label(self, text=self.gather_integrations(), 
                        font=('Arial Bold', 10),
                        bootstyle=ttk.PRIMARY)

        label.pack()

    def gather_integrations(self):
        if not hasattr(self._core, "config"):
            text = "No config loaded"
            return text

        text = self._core.config.filePath.name
        return text





"The device class for the emulator"

import asyncio
import logging
import typing
from typing import *
import random as rnd
import json

from pathlib import Path
from math import ceil
from contextlib import suppress

import tkthread
import tkinter as tk

from PIL import Image, ImageTk

from inkBoarddesigner.platforms.desktop import device
from inkBoard.platforms.basedevice import InkboardDeviceFeatures, FEATURES

from PythonScreenStackManager.devices import PSSMdevice, windowed
from PythonScreenStackManager.tools import DummyTask, TouchEvent
from PythonScreenStackManager import constants as pssmconst

from . import const, pssm_functions

from ..tkinter import window
from ..tkinter.windows import DesignerWindow
from ..tkinter.widgets import PSSMCanvas
from .. import const as des_const

if TYPE_CHECKING:
    from inkBoard import config

_LOGGER = logging.getLogger(__name__)

# from inkBoard.platforms.basedevice import BaseDevice
##Cannot import the device yet -> it messes up the import of pssm
##But, can it not simply be imported later by inkBoard itself?

##Have platforms keep a emulate.json file.
##Holds default values like height and screen type
##Give key for config_features
##Models give keys for stuff

time_hint = Union[int,float,str]

test_init = {
    "required": {
        "model": {"type_hint": 'str'},
        "input": {},
        "stop_time": {"type_hint": str(time_hint)}
    },
    "optional": {

    }
}

def validate_platform_config(emulator_config: dict, config: "config"):
    """Validates the device entry in a config file against the requested platform.

    Parameters
    ----------
    emulator_config : Path
        Configuration settings for the emulator
    config : config
        The config to check

    Raises
    ------
    AssertionError
        Raised when a check fails. Either when the config contains superfluous arguments (aside from platform and name, which are filtered out), or an argument's value does not match the provided type hint.
    """

    platform_conf = emulator_config

    if "__init__" not in platform_conf:
        _LOGGER.warning(f"Cannot validate device config, .json has no __init__ entry.")
        return

    init = platform_conf["__init__"]
    if not "required" in init and not "optional" in init:
        _LOGGER.error("Cannot validate device config, required and/or optional argument keys are not present")
        return

    device_conf = dict(config.device)
    platform_name = device_conf.pop("platform")
    device_conf.pop("name") ##name and platform will be standard. Any platform should accept the name parameter.

    err = False

    for entry in device_conf:
        if entry not in init["required"] and entry not in init["optional"]:
            _LOGGER.error(f"Device config entry {entry} is not a valid argument for platform {platform_name}")
            err = True

    for arg, arg_dict in init["required"].items():
        try:
            assert arg in device_conf, f"Device config is missing required entry {arg}"
            if "type_hint" in arg_dict:
                try:
                    t = eval(arg_dict["type_hint"])
                    if isinstance(t, (type,typing._UnionGenericAlias)):
                        if isinstance(t, typing._UnionGenericAlias):
                            t = t.__args__
                        assert isinstance(device_conf[arg],t), f"Device config entry {arg} is of incorrect type. Should be of type(s) {t}"
                    else:
                        raise NameError(f"{t} is not an instance of type")
                except (NameError, TypeError):
                    _LOGGER.warning(f"Cannot determine type of argument {arg} from string value {arg_dict['type_hint']}. Not validating")
        except (AssertionError, KeyError) as exce:
            _LOGGER.error(exce.args[0], exc_info=None)
            err = True

    for arg, arg_dict in init["optional"].items():
        if arg not in device_conf:
            continue
        try:
            if "type_hint" in arg_dict:
                try:
                    t = eval(arg_dict["type_hint"])
                    if isinstance(t, (type,typing._UnionGenericAlias)):
                        if isinstance(t, typing._UnionGenericAlias):
                            t = t.__args__
                        assert isinstance(device_conf[arg],t), f"Device config entry {arg} is of incorrect type. Should be of type {t}"
                    else:
                        raise NameError(f"{t} is not an instance of type or a union")
                except (NameError, TypeError):
                    _LOGGER.warning(f"Cannot determine type of argument {arg} from string value {arg_dict['type_hint']}")
        except (AssertionError, KeyError) as exce:
            _LOGGER.error(exce.args[0], exc_info=None)
            err = True
    
    assert not err, "Device config failed validation. See logs for more info"
    return

class Device(device.Device):

    def __init__(self, config: "config"):
        
        emulated_platform: str = config.device["platform"]

        if "/" in emulated_platform or "\\" in emulated_platform:
            platform_folder = Path(emulated_platform)
        else:
            platform_folder = const.PLATFORM_FOLDER / emulated_platform

        assert platform_folder.exists(), f"Platform {emulated_platform} does not exist or is not installed"

        self.__emulated_platform = emulated_platform
        self.__emulated_platform_folder = platform_folder

        self._long_click_time = 0.5
        self._bound: set[tuple[tk.Widget, str["sequence"], str["function_id"]]] = set()

        self._emulated_platform = emulated_platform

        if "model" in config.device:
            model = config.device["model"]
        else:
            model = f"{emulated_platform.title()} Device"

        self._model = f"Emulated {model}"

        name = config.device.get("name", None)
        if not name: name = "inkBoard Emulator"

        self.window.title(f"inkBoard Designer - {self._model}")

        emulator_conf = {}
        if (platform_folder / "emulator.json").exists():
            with open(platform_folder / "emulator.json") as f:
                emulator_conf = json.load(f)

        if emulated_platform != "emulator" and config.designer.platform_validation:
            if emulator_conf:
                validate_platform_config(emulator_conf, config)
            else:
                _LOGGER.warning(f"Cannot validate config for platform {emulated_platform}, no emulator.json file")

        self._canvasLock = asyncio.Lock()

        ##These should be checked and be settable by reading out the config
        
        device_map = const.DEFAULT_DEVICE_SCHEMA.copy()
        device_map.update(emulator_conf)

        self._device_map = device_map

        if "width" in config.device:    ##The validation part should take care of width being illegal for a platform
            width = config.device["width"]
        elif device_map["width"] != None:
            width = device_map["width"]
        else:
            width = self.canvas.winfo_width()

        if "height" in config.device:    ##The validation part should take care of width being illegal for a platform
            height = config.device["height"]
        elif device_map["height"] != None:
            height = device_map["height"]
        else:
            height = self.canvas.winfo_height()

        features = InkboardDeviceFeatures(**device_map["features"])

        self.refresh_rate = device_map["refresh_rate"]

        if "screen_mode" in device_map:
            screenMode = device_map["screenmode"]
        elif "screen_type" in device_map:
            screentype = device_map["screen_type"]
            self._screenType = screentype
            if screentype in const.SCREEN_TYPES:
                screenMode = const.SCREEN_TYPES[screentype]
            else:
                screenMode = const.SCREEN_TYPES["default"]

        imgMode = device_map.get("img_mode", f"{screenMode}A")

        defaultColor = device_map.get("defaultColor",None)#"white")

        #Cannot use BaseDevice for this, since that would actually call the wrong super() when initialising
        PSSMdevice.__init__(self,features,
                            width,height,width,height,
                            screenMode, imgMode, defaultColor, name = name)

        if self.has_feature(FEATURES.FEATURE_BATTERY):
            self._battery = Battery(self)
        
        if self.has_feature(FEATURES.FEATURE_BACKLIGHT):
            self._backlight = Backlight(self)

        if self.has_feature(FEATURES.FEATURE_NETWORK):
            self._network = windowed.Network(self)

        self.canvas.set_size(width, height)

        self._canvasWidth = width
        self._canvasHeight = height

        self._windowWidth = self.window.winfo_width()
        self._windowHeight = self.window.winfo_height()

        self._screenImage = Image.new(screenMode,(self.screenWidth,self.screenHeight),
                                    None)
        self.last_printed_PIL = self._screenImage.copy()
        ImageTk.PhotoImage(self.last_printed_PIL)
        self._canvasImageTk = self.window.call_in_main_thread(
                ImageTk.PhotoImage, self.last_printed_PIL)
        
        self._canvasImageTag = None
        self.setup_emulator(config)        
        return
    
    #region
    @property
    def emulated_platform(self) -> str:
        "The platform currently being emulated"
        return self.__emulated_platform
    
    @property
    def emulated_platform_folder(self) -> Path:
        "The path to the actual platform's module"
        return self.__emulated_platform_folder

    @property
    def window(self) -> DesignerWindow:
        "The full designer window"
        return window

    @property
    def canvas(self) -> PSSMCanvas:
        "The tkinter canvas widget that displays the PSSM screen image."
        return window.screenCanvas

    @property
    def screenImage(self) -> Image.Image:
        "The actual image pictured on the screen, as gotten from PSSM. (I.e. the stack)"
        ##Check if this one is actually used in the windowed version
        ##Yes it is, for keeping the screen image, without any alterations from the printTK function
        return self._screenImage
    
    @property
    def screenWidth(self) -> int:
        "Width of the screen"
        return self._canvasWidth
    
    @property
    def viewWidth(self) -> int:
        return self.screenWidth

    @property
    def screenHeight(self) -> int:
        "Height of the screen"
        return self._canvasHeight
    
    @property
    def viewHeight(self) -> int:
        "Height of the screen"
        return self.screenHeight
    #endregion

    def setup_emulator(self, config): 
        ##Meant to setup a session to rougly emulate the platform that inkBoard will run on
        ##The emulator will setup most things itself via the device manifest, however an additional function will be made/checked for to set up additional stuff.
        ##However in that case do not forget to somehow clear those when reloading the config
        
        
        if self.has_feature(FEATURES.FEATURE_INTERACTIVE):
            
            if self.has_feature(FEATURES.FEATURE_PRESS_RELEASE):
                bind_func = self.canvas_event
            else:
                bind_func = self.simple_canvas_event

            funcid = self.canvas.bind("<Button-1>", bind_func)
            self._bound.add((self.canvas, "<Button-1>", funcid))

            funcid = self.canvas.bind("<ButtonRelease-1>", bind_func)
            self._bound.add((self.canvas,"<ButtonRelease-1>", funcid))

        if self.has_feature(FEATURES.FEATURE_RESIZE):
            self._resizeTask = DummyTask()
            funcid = self.window.bind("<Configure>", self._window_configure, add="+")
            self._bound.add((self.window, "<Configure>", funcid))
            if self.window.wm_resizable(None,None) != (True, True):
                self.window.resizable(True, True) ##Doing this reopens the mainwindow, so only doing it if required.
        else:
            if self.window.wm_resizable(None,None) != (False, False):
                self.window.resizable(False, False)

        if self.has_feature(FEATURES.FEATURE_BACKLIGHT):
            self.backlight.set_tkinter_settings()
        return

    def _quit(self, exce):
        for widget, seq, funcid in self._bound:
            try:
                widget.unbind(seq,funcid)
            except tk.TclError:
                continue
        return

    async def _update_canvas(self, img: Image.Image = None):
        """Called by the base windowed device print_pil to update the canvas. Asyncio Lock is implemented to emulate device frame rates."""
        img = img.copy()
        async with self._canvasLock:
            await asyncio.sleep(1/self.refresh_rate)
            tkthread.call_nosync(self.__print_on_canvas, img)
        return

    def __print_on_canvas(self, img):
        self._canvasImageTk = ImageTk.PhotoImage(img)
        if self._canvasImageTag:
            self.canvas.itemconfig(self._canvasImageTag, image = self._canvasImageTk)
        else:
            self._canvasImageTag = self.canvas.create_image(0,0, anchor=tk.NW, image=self._canvasImageTk,
                                    tag=des_const.SCREEN_TAG) 
        self.canvas.update()

    def simple_canvas_event(self, event: tk.Event):
        self._lastEvent = event
        self._interactEvent.set()

    async def simple_canvas_event_handler(self):
        while self.Screen.printing:
            with suppress(asyncio.CancelledError):
                await self._interactEvent.wait()
                self._interactEvent.clear()

                try:
                    await asyncio.wait_for(self._interactEvent.wait(),
                            timeout=self._long_click_time)
                except asyncio.TimeoutError:
                    t = pssmconst.TOUCH_LONG
                    await self._interactEvent.wait()
                else:
                    t = pssmconst.TOUCH_TAP

                self._interactEvent.clear()
                x, y = (self._lastEvent.x, self._lastEvent.y)
                _LOGGER.verbose(f"Passing touch at {(x,y)} as {t}")
                await self.eventQueue.put(TouchEvent(x,y,t))


    async def event_bindings(self, eventQueue = None, grabInput=False):
        pssm_functions.build_element_tree(self.Screen)

        if self.has_feature(FEATURES.FEATURE_INTERACTIVE):
            self._eventQueue = eventQueue
            self._interactEvent = asyncio.Event(loop=self.Screen.mainLoop)
            self.Screen.mainLoop.create_task(self.simple_canvas_event_handler())
        
        self._updateWindowTask = asyncio.create_task(self._update_canvas_loop())

    async def _update_canvas_loop(self):

        while True:
            try:
                await asyncio.sleep(1/des_const.REFRESH_RATE)

            except asyncio.CancelledError:
                return

    async def _resize_window(self, event : Union[tk.Event,asyncio.Event]):
        ##Called when the window has been resized

        if event == None:
            pass

        elif isinstance(event,asyncio.Event):
            _, pending = await asyncio.wait([event.wait()],
                                            timeout=0.25)
            if pending:
                event.set()
            self.window.unbind("<ButtonRelease-1>")
        else:
            return
        
        self._windowWidth = self.window.winfo_width()
        self._windowHeight = self.window.winfo_height()

        self._canvasWidth = self._windowWidth - des_const.INTERFACE_WIDTH
        self._canvasHeight = self._windowHeight

        await self.parentPSSMScreen._screen_resized()

        self._screenImage = Image.new(self.screenMode,(self.screenWidth,self.screenHeight),None)
        self.last_printed_PIL = self._screenImage.copy()

        self.canvas["width"] = self.screenWidth
        self.canvas["height"] = self.screenHeight

        if self.has_feature(FEATURES.FEATURE_BACKLIGHT):
            self.backlight.size = (self.screenWidth,self.screenHeight)
            self.backlight._backlightImg = Image.new("RGBA", color=(0,0,0,0), size=(self.screenWidth,self.screenHeight))

        ##Maybe unlock the lock rn? Or forcibly print it -> do that.
        await self.parentPSSMScreen.print_stack()
        tkthread.call_nosync(self.__print_on_canvas, self.last_printed_PIL.copy())
        return

    def power_off(self, *args):
        if not self.has_feature(FEATURES.FEATURE_POWER):
            _LOGGER.error(f"Platform {self.emulated_platform} does not support the power feature")
        _LOGGER.info("This would have powered off the device")

    def reboot(self, *args):
        if not self.has_feature(FEATURES.FEATURE_POWER):
            _LOGGER.error(f"Platform {self.emulated_platform} does not support the power feature")
        _LOGGER.info("This would have rebooted the device")

class Battery(device.BaseBattery):
    
    randomise = True

    def __init__(self, device):
        
        self._randomise = True

        charge, state = self.update_battery_state()

        super().__init__(device, charge, state)

    async def async_update_battery_state(self)-> tuple[int,str]:
        """
        Update the battery state. Returns the result.

        Returns
        -------
        tuple[int,str]
            Tuple with [charge percentage, state]
        """
        ##idk if this is blocking, if so, needs to go to a thread
        return self.update_battery_state()
    
    def update_battery_state(self):

        if self._randomise:
            self._batteryCharge = rnd.randint(0,100)
            _LOGGER.verbose(f"Emulator has no battery, returning randomised value {self._batteryCharge}%")
            if self._batteryCharge == 100:
                self._batteryState = "full"
            else:
                self._batteryState = rnd.choice(["charging","discharging"])
        else:
            _LOGGER.verbose("Emulator has no battery, not updating")
        return (self._batteryCharge, self._batteryState)
    
    def _set_state(self, value :Literal["full","charging","discharging"]):
        "Manually set the battery state (overwritten when randomise is on and update_state is called)"
        self._batteryState = value

    def _set_charge(self, value : int):
        "Manually set the battery state (overwritten when randomise is on and update_state is called)"
        self._batteryCharge = value

class Backlight(windowed.Backlight):
    "Emulator backlight. Does nothing (may test with putting a dark shaded rectangle over the screen though)"
    def __init__(self, device):
        super().__init__(device)
        self._level = 0
        self.simulate = False
        self.__maxAlpha = 175
        

    @property
    def backlightImage(self) -> Image.Image:
        "The image to simulate the backlight with. Always returns a copy."
        return self._backlightImg.copy()

    def set_tkinter_settings(self, state=False):
        "Sets objects etc. so the backlight can be simulated in the emulator"
        self.screenCanvas = window.screenCanvas
        self.simulate = state
        self.size = (self._device.screenWidth, self._device.screenHeight)
        self.bgRect = False
        self.toggle_simulate(state=state)
        ##Probably do not need to do anything with the hiding/showing here, that can be done in the settings trace.

    def toggle_simulate(self, state : bool = None):
        "Toggles whether the backlight overlay is shown."
        if state == None:
            self.simulate = not self.simulate
        else:
            self.simulate = state
        _LOGGER.debug("Changing backlight")
        
        if self.simulate:
            _LOGGER.debug(f"Toggling backlight simulate. Brt is {self.brightness}")
            alpha = int(self.__maxAlpha - self.__maxAlpha*(self.brightness/100))
            self._backlightImg = Image.new("RGBA", color=(0,0,0,0), size=(self.size[0],self.size[1]))
            blImg = self.backlightImage
            blImg.putalpha(alpha) 
            self.blTk = ImageTk.PhotoImage(blImg)
            self.bgRect = self.screenCanvas.create_image(0,0, anchor=tk.NW, image=self.blTk)
        else:
            if self.bgRect: self.screenCanvas.delete(self.bgRect)

    async def __set_backlight_level(self, level):
        """
        Args:
            level (int): A frontlight level between 0 (off) and 100 (maximum)
        """
        if level < 0 or level > 100:
            return
        
        if not self.simulate:
            return
        
        if level == self._level:
            return
        
        alpha = int(self.__maxAlpha - self.__maxAlpha*(level/100))
        _LOGGER.verbose(f"Backlight brightness to {level}%; Alpha channel is {alpha}")
        blImg = self.backlightImage
        blImg.putalpha(alpha) 
        blTk = ImageTk.PhotoImage(blImg)

        if self.bgRect: 
            self.screenCanvas.itemconfig(self.bgRect, image = blTk)
        else:
            self.bgRect = self.screenCanvas.create_image(0,0, anchor=tk.NW, image=blTk)
        
        self.blTk = blTk
        self._level = level

    async def __transition(self,brightness : int, transition: float):
        if not self.transitionTask.done():
            self.transitionTask.cancel("New transition received")

        self.transitionTask = asyncio.create_task( self.__async_transition(brightness, transition))
        try:
            await self.transitionTask #@IgnoreException
        except asyncio.CancelledError as exce:
            _LOGGER.debug(f"Transition task to {brightness}% in {transition} seconds was cancelled")
        if self._device.parentPSSMScreen.printing:
            async with self._updateCondition:
                self._updateCondition.notify_all()

    async def __async_transition(self, brightness : int, transition: float):
        """
        Async function to provide support for transitions. Does NOT perform sanity checks

        Parameters
        ----------
        brightness : int
            the desired end brightness
        transition : float
            the transition time in seconds
        """
                    
        if transition == 0:
            await self.__set_backlight_level(brightness)
            return

        if self.brightness == brightness:
            return

        min_wait = 0.05
        wait = transition/(abs(self.brightness-brightness))
        step = -1 if self.brightness > brightness else 1

        async with self._lightLock:
            if wait < min_wait: 
                steps = ceil(transition/min_wait)
                for i in range(0,steps):
                    step = int(brightness - self.brightness)/(steps-i)
                    L = asyncio.gather(self.__set_backlight_level(self.brightness + step), 
                                    asyncio.sleep(min_wait))
                    await L #@IgnoreException
                await self.__set_backlight_level(brightness)
            else:
                while self.brightness != brightness:
                    L = asyncio.gather(self.__set_backlight_level(self.brightness + step), asyncio.sleep(wait))
                    await L #@IgnoreException

    async def turn_on_async(self, brightness : int = None, transition: float = None):
        """Async function to provide support for transitions at turn on. Does NOT perform sanity checks"""
        if self.simulate:
            _LOGGER.debug(f"Async turning on in {transition} seconds")
        
        if brightness == None:
            brightness = self.defaultBrightness
        
        if transition == None:
            transition = self.defaultTransition

        if self.brightness == brightness:
            ##Do nothing if the light is already at the correct level
            return

        await self.__transition(brightness,transition)

    def turn_on(self, brightness : int = None, transition : float = None):
        """Turn on the backlight to the set level"""

        if transition == None:
            transition = self.defaultTransition

        if brightness == None:
            brightness = self.defaultBrightness
        
        if transition < 0:
            _LOGGER.error("Transition time cannot be negative.")
            return
        
        if brightness < 0 or brightness > 100:
            _LOGGER.error(f"Brightness must be between 0 and 100. {brightness} is an invalid value")
            return
        
        asyncio.create_task(self.turn_on_async(brightness, transition))

    async def turn_off_async(self, transition: float = None):
        """Async function to provide support for transitions at turn off. Does NOT perform sanity checks"""
        _LOGGER.debug("Async turning off")
        if not self.state:
            ##Do nothing if the light is already off
            return

        if transition == None:
            transition = self.defaultTransition

        await self.__transition(0,transition)

    def turn_off(self, transition : float = None):
        """Turns off the backlight to the set level"""
        if not self.state:
            ##Backlight is already off, no need to do anything
            return

        if transition == None:
            transition = self.defaultTransition

        if transition < 0:
            _LOGGER.error("Transition time cannot be negative.")
            return

        asyncio.create_task(self.turn_off_async(transition))

    def toggle(self, brightness : int = None, transition : float = None):
        if self.state:
            self.turn_off(transition)
        else:
            self.turn_on(brightness,transition)

    async def toggle_async(self, brightness: int = None, transition: float = None):
        if self.state:
            await self.turn_off_async(transition)
        else:
            await self.turn_on_async(brightness,transition)




from pathlib import Path

from inkBoard.constants import INKBOARD_FOLDER
from inkBoard.platforms.basedevice import FEATURES

DEFAULT_DEVICE_SCHEMA = {
    "screen_type": "LED",
    "width": None,
    "height": None,
    "refresh_rate": 20,
    "features": {
        FEATURES.FEATURE_NETWORK: True,
        FEATURES.FEATURE_INTERACTIVE: True,
        FEATURES.FEATURE_ROTATION: False,
        FEATURES.FEATURE_RESIZE: True,
        FEATURES.FEATURE_BATTERY: True,
        FEATURES.FEATURE_BACKLIGHT: True,
        FEATURES.FEATURE_PRESS_RELEASE: True,
        FEATURES.FEATURE_POWER: False
    }
}

SCREEN_TYPES = {
    "default": "RGB",
    "e-ink": "L"
}

PLATFORM_FOLDER = Path(__file__).parent.parent / "platforms"
"Folder holding the available platforms."


"""
This library manages elements from PSSM and linked entities from Home Assistant, and provides additional comound elements in the HAelements library.
Import it first, or at least before importing the element library to allow it to add to the PSSM default colors and fonts.
"""

from typing import TYPE_CHECKING, Optional
from types import MappingProxyType
import logging

from inkBoard.helpers import *
from PythonScreenStackManager import constants as pssm_const

from . import constants as const

if TYPE_CHECKING:
    from inkBoard import config, core as CORE
    from mdi_pil import mdiType
    from PythonScreenStackManager import pssm_types as pssm
    from PythonScreenStackManager.pssm.screen import PSSMScreen
    from . import client

logger = logging.getLogger(__name__)


##This one is required
async def async_setup(core: "CORE", config : "config") -> None:
    ##Screen can be gotten by importing at this point

    ha_config = config.configuration["home_assistant"]

    if "url" not in ha_config or "token" not in ha_config:
        logger.error("Home Assistant integration requires url and token to be set in the config. At least one is missing.")
        return False

    from . import client
    client.CORE = core
    HAclient = client.HAclient(core.screen)
    from . import parser
    parser.setup_elements()

    core.add_element_parser("HA", parser.parse_ha_element)
    pssm_const.PSSM_COLORS['home-assistant'] = const.HOMEASSISTANT_BLUE
    pssm_const.PSSM_COLORS['homeassistant'] = const.HOMEASSISTANT_BLUE

    pssm_const.SHORTHAND_FONTS['home-assistant'] = 'quicksand-bold' ##The font comes with pssm by default
    pssm_const.SHORTHAND_FONTS['homeassistant'] = 'quicksand-bold'

    pssm_const.SHORTHAND_ICONS['home-assistant'] = const.HOMEASSISTANT_ICON
    pssm_const.SHORTHAND_ICONS['homeassistant'] = const.HOMEASSISTANT_ICON
    return HAclient

async def async_start(core: "CORE", client : "client.HAclient"):
    client.setup_entity_functions()
    await client.connect_client()
    return

class home_assistantMap(TypedDict):
    "Dict with settings required for the home assistant client"

    url : str
    "Url to the home assistant server."

    token: str
    "Long lived access token to authenticate with the server."

    state_colors: dict
    """
    This way you can map the default foreground colors of connected elements to take on the same color when their state matches.
    Also accepts a default entry for unknown states.
    Will be used for elements where state_colors: True is set.
    """

    ping_pong_interval : 'pssm.DurationType'
    "Interval inbetween checking the connection to the server. Generally you can keep this undefined."

    unknown_icon : Optional['mdiType']
    "Default icon to indicate that an entity's state is unknown"

    unavailable_icon : Optional['mdiType']
    "Default icon to indicate that an entity is unavailable"

home_assistantMap.__required_keys__ = frozenset({'url','token'})
home_assistantMapDefaults = MappingProxyType({"state_colors": {}, "ping_pong_interval": 50, 'unknown_icon': "mdi:help", 'unavailable_icon': "mdi:exclamation-thick"})




import asyncio
from typing import *
import tkinter as tk
import logging
from pathlib import Path
from PIL import Image, ImageOps

from mdi_pil import mdiType
from mdi_pil.ttkbootstrap_mdi import PhotoIcon, MDIIcon

import ttkbootstrap as ttk
from ttkbootstrap.style import Bootstyle
from ttkbootstrap.tooltip import ToolTip

from . import functions as tk_functions
from .. import const
from ..util import iidType

if TYPE_CHECKING:
    from .windows import DesignerWindow
    from ..emulator.device import Device

    from PythonScreenStackManager.elements import Element

_LOGGER = logging.getLogger(__package__)
_LOGGER.setLevel(logging.DEBUG)

_LOGGER = logging.getLogger(__package__)
_LOGGER.setLevel(logging.DEBUG)

window: "DesignerWindow"

class LabelToggle(ttk.Labelframe):

    "Labelframes with a toggle widget inside. Use the noframe.TLabelframe style defined in configure_themes"

    def __init__(self, master = None, *, labelanchor = "w", cursor = const.INTERACT_CURSOR, style="noframe.TLabelframe",
                variable: tk.BooleanVar = None, command = "", **kwargs):
        super().__init__(master, cursor=cursor, labelanchor=labelanchor, style=style, **kwargs)        
        
        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.columnconfigure(2,weight=1)

        self.variable: Optional[tk.BooleanVar] = variable
        self._command = command

        b = self.cget("style")
        bootstyle = Bootstyle.ttkstyle_widget_color(b)
        if not bootstyle: bootstyle="primary"

        self.CheckButton = ttk.Checkbutton(self,
                onvalue = 1, offvalue = 0, bootstyle=f"{bootstyle}-round-toggle",
                variable=self.variable, command=command, cursor=cursor)
        self.CheckButton.grid(row=0,column=1, sticky=tk.E)
        self.bind(sequence="<Button-1>",func=self.toggle)


    def toggle(self, event):

        tkvar = self.variable
        if tkvar: tkvar.set(not tkvar.get())
        if self._command: self._command(event)

class LabelIcon(ttk.Labelframe):

    "Label widget with an mdi icon in a button in it."
    
    def __init__(self,  icon: mdiType, icon_size: int = 50, master = None, *, labelanchor = "w", cursor = "",
                command = "", style="noframe.TLabelframe", **kwargs):
        
        super().__init__(master, cursor=cursor, labelanchor=labelanchor, style=style, **kwargs)        
        
        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.columnconfigure(2,weight=1)

        b = self.cget("style")
        bootstyle = Bootstyle.ttkstyle_widget_color(b)
        if not bootstyle: bootstyle="primary"

        self._iconImg = MDIIcon(icon, (icon_size,icon_size), bootstyle=bootstyle)
        
        self._button = ttk.Button(self,image=self._iconImg, cursor=cursor, 
                        width=icon_size, padding=-1, style=f"image.TButton",
                        )
        
        self._button.grid(row=0,column=0, columnspan=3, sticky="nsew")
        if command:
            self.bind("<Button-1>",func=command)
            self._button.bind("<Button-1>",func=command)
        

class Treeview(ttk.Treeview):

    """Treeview with base styling and bindings for hovering.

    Usage is a bit complicated, since just using this as a subclass for some reason caused the Tree to stop opening items when clicking on the + (which also happens when setting up the tree in the __init__).
    So set up a Treeview first, and pass that. The class will function exactly the same as a normal treeview.
    """

    _treeframe: ttk.Frame = None

    def __init__(self, tree: ttk.Treeview, 
                on_select: Callable[["Treeview", tk.Event, tuple[iidType,]], None] = const.DEFAULT,
                on_hover: Callable[["Treeview", tk.Event, iidType], None] = const.DEFAULT,
                on_double_click: Callable[["Treeview", tk.Event, iidType], None] = const.DEFAULT):
        

        self.__tree_ph = tree

        tree.configure(style=const.TREEVIEW_STYLE, selectmode="browse")

        if tree.master != Treeview._treeframe:
            _LOGGER.verbose("Tree has the wrong master widget!")
            tree.master = Treeview._treeframe

        self.tooltip = ToolTip(self.tree, bootstyle=const.TOOLTIP_STYLE)
        self.last_hover: str = None
        "The last item hovered over in this tree"

        if on_select == const.DEFAULT: on_select = self.highlight_element
        if on_hover == const.DEFAULT: on_hover = None
        if on_double_click == const.DEFAULT: on_double_click = None

        self.__on_hover_function = None
        self.__on_select_function = on_select
        self.__on_double_click_function = None

        self._element_iids: dict[str,set] = {}
        #Holds all registered iid for a given element

        self._element_items : dict[str,"Element"] = {}
        #Maps iids to elements

        self.tooltip.move_tip()
        self.tooltip.hide_tip()

        self.tree.tag_configure(const.HOVER_TAG, background="grey")
        self.tree.bind("<Motion>", self._hover)
        self.tree.bind("<Leave>", self._leave)

        self.tree.bind("<Button-1>", self._click)
        self.tree.bind("<<TreeviewSelect>>", self.__on_select)
        self.tree.bind("<Double-1>", self.__on_double_click)


    def __getattr__(self, name):
        return getattr(self.tree, name)

    #region
    @property
    def tree(self) -> ttk.Treeview:
        "The actual treeview"
        return self.__tree_ph

    @property
    def on_hover(self) -> Callable[["Treeview", tk.Event, iidType], None]:
        """Function to call when a new item is hovered over (focused). 
        Useful to configure the text of the tooltip. Defaults to None."""
        return self.__on_hover_function

    @on_hover.setter
    def on_hover(self, value):
        if not isinstance(value, Callable):
            value = None
        self.__on_hover_function = value
        return
    
    @property
    def on_select(self) -> Callable[["Treeview", tk.Event, tuple[iidType,]], None]:
        """Function to call whenever an item is selected. 
        Does not directly pas an iid, but the tuple of selected iid's"""
        return self.__on_select_function
    
    @on_select.setter
    def on_select(self, value):
        if not isinstance(value, Callable):
            value = None
        self.__on_select_function = value

    @property
    def on_double_click(self) -> Callable[["Treeview", tk.Event, iidType], None]:
        "Function to call whenever an item is double clicked. Defaults to None"
        return self.__on_double_click_function
    
    @on_double_click.setter
    def on_double_click(self, value):
        if not isinstance(value, Callable):
            value = None
        self.__on_double_click_function = value
    #endregion

    def _hover(self, event):

        _iid = self.identify_row(event.y)
        if not _iid:
            if self.last_hover:
                self.item(self.last_hover, tags=[])
                self.last_hover = None
            self.tooltip.hide_tip()
            return
        
        if _iid != self.last_hover:
            _LOGGER.verbose(f"Hovered over iid {_iid}")
            if self.last_hover:
                self.tooltip.hide_tip()
                self.item(self.last_hover, tags=[])
            self.item(_iid, tags=[const.HOVER_TAG])
            self.last_hover = _iid
            self.__on_hover(event, _iid)

    def _leave(self, event):    
        self.tooltip.hide_tip()
        if self.last_hover:
            self.tooltip.hide_tip()
            self.item(self.last_hover, tags=[])
            self.last_hover = None
        return
    
    def _click(self, event: tk.Event):
        if not self.identify_row(event.y):
            for item in self.selection():
                self.selection_remove(item)

    def __on_select(self, event: tk.Event):
        if not self.on_select: return
        if not self.selection(): return
        return self.on_select(self, event, self.selection())

    def __on_hover(self, event, iid):
        "Settable, called when a new tree item is hovered over"
        if not self.on_hover: 
            return
        return self.on_hover(self, event, iid)
    
    def __on_double_click(self, event: tk.Event):
        "Settable, called when a new tree item is hovered over"
        if not self.on_double_click: 
            return
        iid = self.identify_row(event.y)
        if not iid: return
        return self.on_double_click(self, event, iid)    
    
    def state(self,statespec=None):
        if statespec:
            e = super().state(statespec)
            if 'disabled' in e:
                self.tree.bindtags(self.tags)
            elif '!disabled' in e:
                self.tags = self.tree.bindtags()
                self.tree.bindtags([None])
            return e
        else:
            return super().state()

    def disable(self):
        self.state(('disabled',))

    def enable(self):
        self.state(('!disabled',))

    def is_disabled(self):
        return 'disabled' in self.state()

    def is_enabled(self):
        return not self.is_disabled()
    
    def insert_element_item(self, element: "Element", text: str, parent_iid="", index=tk.END) -> str:
        """
        Shorthand function to add an item for an element to the tree, such that it can be used with the highlight function.

        Parameters
        ----------
        element : Element
            The element to build the tree item of
        text : str
            String to put as text in the item
        parent_iid : str, optional
            iid of the parent item, by default ""
        index : _type_, optional
            Index to insert into, by default tk.END

        Returns
        -------
        str
            iid of the new item
        """

        if element.id in self._element_iids:
            iid_idx = len(self._element_iids[element.id]) + 1
        else:
            self._element_iids[element.id] = set() ##Using sets to prevent any duplicates
            iid_idx = 0
        
        iid = f"{element.id}-iid_{iid_idx}"
        self._element_iids[element.id].add(iid)
        self._element_items[iid] = element

        item_iid = self.tree.insert(
            parent_iid,
            index=index,
            iid=iid,
            text=text,
            image=tk_functions.get_element_tree_icon(element)
            )
        return item_iid
    
    def highlight_element(self, tree, event, iid):
        
        iid = iid[0]
        if iid not in self._element_items:
            tk_functions.highlight_elements() ##Remove any present highlights
            return
        
        elt = self._element_items[iid]
        tk_functions.highlight_elements(elt)
        return

class CanvasBackground(PhotoIcon):
    "MDI_PIL PhotoIcon implementation to change the background image depending on theme type."

    def __init__(self, size, **kw):

        self._file_light = Path(__file__).parent.parent / "files" / "background_alpha_light.png"
        self._file_dark = Path(__file__).parent.parent / "files" / "background_alpha_dark.png"

        assert self._file_light.exists()

        image = self._create_images(size)

        super().__init__(image, size, **kw)

    def _create_images(self, size):
        self._image_light = ImageOps.pad(Image.open(self._file_light),
                                    size, centering=(1,1))
        self._image_dark = ImageOps.pad(Image.open(self._file_dark),
                                    size, centering=(1,1))
        style = ttk.Style.get_instance()
        self._current_type = style.theme.type
        if self._current_type == ttk.LIGHT:
            return self._image_light
        else:
            return self._image_dark

    def _change_img_style(self, *args):
        
        style = ttk.Style.get_instance()
        if style.theme.type == self._current_type:
            return
        
        img = self._img
        
        if style.theme.type == ttk.LIGHT:
            new_img = self._image_light
        else:
            new_img = self._image_dark

        new_img.putalpha(img.getchannel("A"))
        self.paste(new_img)
        self._current_type = style.theme.type

class PSSMCanvas(tk.Canvas):
    """tkinter Canvas widget that can be specialised for inkBoard designer.
    """
    master: "DesignerWindow"

    def __init__(self, master = None, **kwargs):
        
        self._background_Tk_idx = None
        super().__init__(master, **kwargs)
        
    def set_size(self, width: int, height: int):
        "Sets the size of the Canvas by correctly updating the size of the designer window."

        window_width = width + const.INTERFACE_WIDTH
        window_height = height

        self.master.geometry(f"{window_width}x{window_height}")
        self.master.update()
        self.master.update_idletasks()
        self._build_canvas_background()

    
    def _build_canvas_background(self):

        size = (self.winfo_width(),
                self.winfo_height())
        self._background_Tk = CanvasBackground(size)

        if self._background_Tk_idx:
            self.itemconfig(self._background_Tk_idx, image =  self._background_Tk)
        else:
            self._background_Tk_idx = self.create_image(0,0, anchor=tk.NW, image=self._background_Tk)
        self.update()
        return

    def _clear(self):
        "Clears the screen canvas from everything except the background"
        for idx in self.find_all():
            if idx == self._background_Tk_idx: 
                continue
            ##May need to check if this would cause issues with the rectangles, as it would obviously create a mismatch between them
            self.delete(idx)
        self.unbind("<Button>")
        self._build_canvas_background()
        self.update()



class BatteryFrame(ttk.Labelframe):
    "Frame to configure emulator battery settings"

    def __init__(self, master, device: "Device", **kwargs):

        super().__init__(master, text="Battery Settings",  
                        bootstyle=const.FEATURE_FRAME_STYLE, **kwargs)

        self._device = device

        frame = ttk.Frame(self)
        frame.grid(column=0, row=0, sticky="nsew",pady=(10,10))

        var = tk.BooleanVar(self,name="device.battery_rnd", value=device.battery._randomise)
        toggleFrame = LabelToggle(frame, text="Randomise", variable=var, command=self.__toggle_rnd)
        toggleFrame.grid(row=1, column=0, sticky="nsew", pady=(10,0))
        ToolTip(toggleFrame, "Return a random battery value and state each time the value is requested.",  bootstyle=const.TOOLTIP_STYLE)
        
        stateVar = ttk.StringVar(master=self, name="battery.state")    
        stateVar.trace_add("write",self.__set_state)

        ops = ["full","charging","discharging"]
        stateFrame = ttk.Labelframe(frame,text="State", style='toggleCenter.TLabelframe', labelanchor="n")
        stateBox = ttk.Combobox(stateFrame, textvariable=stateVar, values=ops, state="readonly", width=25, cursor="hand2")
        
        stateBox.grid(row=0,column=0)
        stateFrame.grid(row=0,column=0)


        digit_func = device.window.register(self.__validate_charge)
        chargevar = ttk.IntVar(self, device.battery.charge, "battery.charge")
        chargevar.trace_add("write",self.__set_charge_var)
        trFrame = ttk.Labelframe(frame,text="Charge", style='toggleCenter.TLabelframe', labelanchor="n")
        trFrame.columnconfigure(0,weight=1)
        chargeBox = ttk.Spinbox(trFrame, from_=0, to=100, command=self.__set_charge, 
                                    validate="key", validatecommand=(digit_func, '%P', "%W"))
        chargeBox.grid(row=0,column=0, sticky="ew")
        trFrame.grid(row=0, column=1)

        button = ttk.Button(frame, command=self.__call_update, text="Update Now", style="primary")
        button.grid(row=1, column=1)
        ToolTip(button,)

        self.chargevar = chargevar
        self.chargeBox = chargeBox
        self.stateVar = stateVar

        chargeBox.set(chargevar.get())
        stateBox.set(self._device.battery.state)

    def __toggle_rnd(self, *args):
        val = self.getvar("device.battery_rnd")
        
        self._device.battery._randomise = bool(val)

    def __validate_charge(self, value, widget_name):
        try:
            value = int(value)
        except ValueError:
            return False
            
        w = self.nametowidget(widget_name)

        if value >= 0 or value <= 100:
            w.set(value)
            ##This retrieves the actual command attached to the widget
            cmd = lambda : self.tk.call(w['command'])
            cmd()
            return True
        else:
            return False
        
    def __set_state(self, *args):

        val = self.stateVar.get()

        if val == "full" and self._device.battery.charge != 100:
            _LOGGER.warning("Battery only reports full at 100%, change the charge first")
        self._device.battery._set_state(val)

    def __set_charge_var(self, var_name, *args):
        val = self.chargevar.get()
        self._device.battery._set_charge(val)
        return

    def __set_charge(self, *args):
        val = self.chargeBox.getint(self.chargeBox.get())
        self._device.battery._set_charge(val)
        return

    def __call_update(self, *args):
        async def notify_condition():
            condition = self._device.parentPSSMScreen.deviceUpdateCondition
            async with condition:
                condition.notify_all()

        asyncio.create_task(notify_condition())


class BacklightFrame(ttk.Labelframe):
    "Frame to configure emulator backlight settings"

    def __init__(self, master, device: "Device", **kwargs):

        self._device = device

        super().__init__(master, text="Backlight Settings",  
                        bootstyle=const.FEATURE_FRAME_STYLE, **kwargs)

        frame = ttk.Frame(self)
        frame.grid(column=0, row=0, sticky="nsew")

        num_cols = 5
        slider_pad = (5,5)

        self.simulate_var = tk.BooleanVar(device.window, name="backlight.simulate")
        toggleFrame = LabelToggle(master=frame, labelanchor=tk.N, text="Simulate", variable=self.simulate_var, command=self.__toggle_backlight)
        toggleFrame.grid(row=0, column=num_cols-1, sticky="nsew")
        ToolTip(toggleFrame, "Place a semi-transparent dark rectangle over the screen, to simulate the backlight. Rectangle becomes more transparent when brightness goes up.",  bootstyle=const.TOOLTIP_STYLE)

        val = device.Screen.backlight_behaviour.title()
        self.backlightBehv = ttk.StringVar(master=device.window, name="backlight.behaviour")
        self.backlightBehv.trace_add("write",self.__set_behaviour)
                
        ops = ["Manual", "On Interact", "Always"]
        behvlabel = ttk.Label(frame, text="Behaviour", anchor=tk.W)
        behv = ttk.Combobox(frame, textvariable=self.backlightBehv, values=ops, state="readonly", width=25, cursor="hand2")
        behv.set(val)
        behvlabel.grid(row=0, column=0, sticky=tk.W)
        behv.grid(row=0, column=1, columnspan=num_cols-2)
        ToolTip(behvlabel,"Backlight behaviour. Manual means it only turns on/off when called explicitly. On Interact means it turns on when interacting with the screen. Always means it is always on by default (can be turned on/off via functions). Initial behaviour is set in configuration.yaml", bootstyle=const.TOOLTIP_STYLE)


        brtLabel = ttk.Label(frame, text="Brightness", anchor=tk.W)
        var = tk.IntVar(self, name="backlight.brightness", value=device.backlight.brightness)
        brtSlide = ttk.Scale(frame, command=self.__set_brightness, name="brightness",  from_=0, to=100, length=None, cursor="hand2", variable=var)
        brtSlide.grid(row=1,column=1, sticky=tk.EW, columnspan=num_cols-1, pady=slider_pad)
        brtLabel.grid(row=1, column=0, sticky=tk.EW, padx=(0,10))
        brtSlide.bind('<Button-1>', tk_functions.set_slider_left_click)


        defbrtLabel = ttk.Label(frame, text="Default Brightness", anchor=tk.W)
        defbrtSlide = ttk.Scale(frame, command=self.__set_default_brightness, name="default_brightness",  from_=0, to=100, length=None, cursor="hand2")
        defbrtSlide.grid(row=2,column=1, sticky=tk.EW,  columnspan=num_cols-1, pady=slider_pad)
        defbrtLabel.grid(row=2, column=0, sticky=tk.EW, padx=(0,10))
        defbrtSlide.bind('<Button-1>', tk_functions.set_slider_left_click)
        ToolTip(defbrtLabel, "The default brightness (0-100) when turning on the backlight without specifying a brightness.", bootstyle=const.TOOLTIP_STYLE)

        ##Spinboxes
        spFrame = ttk.Frame(frame)
        spFrame.grid(row=3, columnspan=num_cols)
        digit_func = window.register(tk_functions.validate_positive_number)

        ##Default transition time
        trFrame = ttk.Labelframe(spFrame,text="Transition Time", style='toggleCenter.TLabelframe', labelanchor="n")
        trFrame.columnconfigure(0,weight=1)
        self.trBox = ttk.Spinbox(trFrame, command=self.__set_default_transition , validate="all", validatecommand=(digit_func, '%P', "%W"),
                                from_=0.0, to=60*60, increment=1) ##Somehow increments smaller than one mess up the widget so I guess it'll be like this.  
        self.trBox.set(device.backlight.defaultTransition)
        self.trBox.grid(row=0,column=0, sticky="ew")
        trFrame.grid(row=0, column=0)
        ToolTip(trFrame, "The default transition time (in seconds) when turning on/off the backlight without specifying a transition.", bootstyle=const.TOOLTIP_STYLE)
    
        offFrame = ttk.Labelframe(spFrame,text="Default on time", style='toggleCenter.TLabelframe', labelanchor="n")
        offFrame.columnconfigure(0,weight=1)
        self.offBox = ttk.Spinbox(offFrame, from_=0.0, to=60*60, command=self.__set_turn_off_time ,validate="key", validatecommand=(digit_func, '%P', "%W"))
        self.offBox.set(device.parentPSSMScreen.backlight_time_on)
        self.offBox.grid(row=0,column=0, sticky="ew")
        offFrame.grid(row=0, column=1)
        ToolTip(offFrame, "The default time the backlight stays on for (in seconds) when calling the screen's temporary backlight function without specifying a time.", bootstyle=const.TOOLTIP_STYLE)


    def __set_behaviour(self, *args):
        n = self.backlightBehv.get()
        self._device.Screen.set_backlight_behaviour(n)

    def __set_brightness(self, val):
        val = float(val)
        self._device.backlight.turn_on(int(val), transition=0)

    def __set_default_brightness(self, event: tk.Event):
        val = float(event)
        self._device.backlight.defaultBrightness = int(val)

    def __toggle_backlight(self, *args):
        var_name = "backlight.simulate"
        val = self.getvar(var_name)
        self._device.backlight.toggle_simulate(val)
        return

    def __set_default_transition(self,*args):
        val = self.trBox.get()
        self._device.backlight.defaultTransition = float(val)
        return

    def __set_turn_off_time(self,*args):
        val = self.offBox.get()
        self._device.parentPSSMScreen.backlight_time_on = float(val)






"Runners for various important functions of the designer."

import tkthread
tkthread.patch()

from typing import TYPE_CHECKING
import inkBoard.packaging

import asyncio
import concurrent.futures
import threading

import sys
from pathlib import Path

import ttkbootstrap as ttk

import inkBoard
from inkBoard import constants as ib_const, core as CORE
from inkBoard.helpers import QuitInkboard, ConfigError, DashboardError, DeviceError, ScreenError

from . import util, const, _LOGGER
from .settings import save_settings

from .tkinter import window, functions as tk_functions
from .tkinter.builders import build_window
from .tkinter.functions import stop_emulator

if TYPE_CHECKING:
    import inkBoard.bootstrap
    import inkBoard.constants
    import inkBoard.helpers

importer_thread = concurrent.futures.ThreadPoolExecutor(None,const.IMPORTER_THREADPOOL)


def stop_designer():
    window._mainLoop.create_task(async_stop_designer())

async def async_stop_designer():
    save_settings()
    if hasattr(CORE,"screen"):
        try:
            stop_emulator(exce=SystemExit("Closing Designer"))
        except:
            pass
    
    if window._inkBoard_thread:
        window._inkBoard_thread.join(timeout=5)
    await asyncio.sleep(0)
    for task in asyncio.all_tasks():
        task.cancel()
    window.stop_update_loop()
    await asyncio.sleep(0)
    return

async def async_unload_inkBoard(reload_modules: bool = False):
    "Cancels all tasks running in the PSSM loop and reloads necessary modules."
    
    _LOGGER.debug(f"Unloading inkBoard")
    stop_loop = False
    if hasattr(CORE,"screen") and CORE.screen.mainLoop:
        CORE.screen.mainLoop.stop()
        stop_loop = CORE.screen.mainLoop
    
    await asyncio.to_thread(window._inkBoard_lock.acquire)
    await asyncio.to_thread(window._inkBoard_thread.join)

    if not hasattr(inkBoard, "core"):
        window._inkBoard_clean = True
        window._inkBoard_lock.release()
        return

    await asyncio.sleep(0)

    if hasattr(CORE,"screen"):
        for task in asyncio.all_tasks(CORE.screen.mainLoop):
            task.cancel()
    
    await asyncio.sleep(0)

    if reload_modules:
        ##Reloading the platform may actually be done already via the platform check
        await asyncio.to_thread(
            inkBoard.bootstrap.reload_core, CORE, True
                )
    else:
        await asyncio.to_thread(
            inkBoard.bootstrap.reload_core, CORE, False
                )
        ##Regarding that: move shorthand colors to style, so they can be reset from there.

    if stop_loop:
        stop_loop.close()

    window._inkBoard_clean = True
    window._inkBoard_lock.release()
    return

def unload_inkBoard(reload_modules: bool = False):
    t = window._mainLoop.create_task(async_unload_inkBoard(reload_modules))
    return t

async def reload_config(config, full_reload: bool = False):
    await unload_inkBoard(True)
    window.set_progress_bar(value=-1)
    window._mainLoop.create_task(run_inkboard_config(config))

async def run_inkboard_thread(config_file):
    
    reload_finally = True
    try:
        config_path = Path(config_file)
        config_folder = config_path.parent.absolute()
        custom_folder = config_folder / "custom"

        if window._inkBoard_lock.locked():
            _LOGGER.warning("Attempting to run new inkBoard thread before the last one has fully shut down.")

        window.set_progress_bar(1, text="Acquiring resources", title=f"Loading {config_path.name}")
        tkthread.call_nosync(window.configLabel.configure, 
                            text = config_path.name, cursor=const.INTERACT_CURSOR)

        window._inkBoard_lock.acquire()
        window._inkBoard_clean = False

        window.set_progress_bar(5, "Importing base functions")

        from inkBoard import core as CORE

        from .emulator import pssm_functions
        pssm_functions.CORE = CORE
        tk_functions.CORE = CORE

        _LOGGER.debug(f"CORE imported at {CORE.IMPORT_TIME}")

        from inkBoard import bootstrap 

        from .integrationloader import IntegrationLoader
        
        CORE.integration_loader = IntegrationLoader

        window.set_progress_bar(value=10, text="Gathering available integrations")
        folders = {"custom.integrations": custom_folder / "integrations"} | const.INTEGRATION_DIRS

        IntegrationLoader.get_integrations(folders)

        window.set_progress_bar(value=20, text="Importing PythonScreenStackManager")

        from PythonScreenStackManager.exceptions import ReloadWarning, FullReloadWarning

        window.set_progress_bar(value=25, text="Reading out base config")

        CORE.config = bootstrap.setup_base_config(config_file)

        bootstrap.setup_logging(CORE)

        window.set_progress_bar(30, "Importing integrations")
        CORE.integration_loader.import_integrations(CORE,window.set_progress_bar,(30,42))

        ##Is there a reason to not do this after setting up the screen and stuff?
        ##Except disallowing defining the screen outside of core
        CORE.custom_functions = bootstrap.import_custom_functions(CORE)

        bootstrap.import_custom_elements(CORE)

        ##Implement error catchers for these as well
        window.set_progress_bar(42, "Setting up emulator device")
        CORE.device = await bootstrap.setup_device(CORE)

        window.set_progress_bar(45, "Setting up PythonScreenStackManager screen")
        CORE.screen = await bootstrap.setup_screen(CORE)
        screen = CORE.screen
        screen.add_shorthand_function_group("custom", CORE.parse_custom_function)

        window.set_progress_bar(50, "Setting up styles")
        bootstrap.setup_styles(CORE)

        window.set_progress_bar(52, "Setting up integrations")
        max_integration_progress = 70
        CORE.integration_objects = await IntegrationLoader.async_setup_integrations(CORE, window.set_progress_bar, 
                                                                                (window._progressBar["value"],max_integration_progress))

        window.set_progress_bar(max_integration_progress, "Setting up dashboard")
        main_layout = bootstrap.setup_dashboard_config(CORE)

        window.set_progress_bar(75, "Readying screen")
        screen.clear()
        screen.start_batch_writing()

        window.set_progress_bar(77, "Starting integrations")
        await IntegrationLoader.async_start_integrations(CORE)

        window.set_progress_bar(90, "Preparing screen and elements for printing")
        await screen.async_add_element(main_layout, skipPrint=True)
        screen.stop_batch_writing()

        window.set_progress_bar(95, "Starting printing")

        try:
            done, pending = await asyncio.wait([bootstrap.run_core(CORE)], timeout=0.5)
            if pending:
                window.set_inkboard_state(ttk.ACTIVE)
                window.set_progress_bar(100, "Up and running")
                await asyncio.gather(*pending)
            else:
                for t in done:
                    _LOGGER.error(t.result())
                await asyncio.sleep(0)
                raise RuntimeWarning            
        except asyncio.CancelledError:
            window.set_inkboard_state(ttk.DISABLED)
        except ReloadWarning as exce:
            window.set_inkboard_state(ttk.DISABLED)
            if exce == FullReloadWarning or isinstance(exce, FullReloadWarning):
                full_reload = True
            else:
                full_reload = False
            window._mainLoop.create_task(reload_config(config_file, full_reload))
            reload_finally = False
        except QuitInkboard:
            window.set_inkboard_state(None)
            await asyncio.sleep(0)
        except SystemExit:
            window._inkBoard_lock.release() #Keeping this here to catch out the errors in IDE
            reload_finally = False
        except KeyboardInterrupt:
            stop_designer()
            
        except RuntimeError:
            pass
    except DashboardError as exce:
        CORE.screen.mainLoop.stop()
        _LOGGER.error(f"Error in dashboard config of file {config_file}: {exce}")
        window.set_inkboard_state("ERROR")
        window.set_progress_bar(ttk.DANGER, f"Error setting up dashboard: {exce}")
        await asyncio.sleep(0)
    except ConfigError as exce:
        _LOGGER.error(f"Error in config file {config_file}: {exce}")
        window.set_inkboard_state("ERROR")
        window.set_progress_bar(value=ttk.DANGER, text=f"Error in config file {config_file}: {exce}")
    except DeviceError as exce:
        msg = f"Error setting up inkBoard device: {exce}"
        _LOGGER.error(msg, exc_info=None)
        _LOGGER.debug(f"{type(exce)} info:", exc_info=exce)
        window.set_inkboard_state("ERROR")
        window.set_progress_bar(value=ttk.DANGER, text=msg)
    except ScreenError as exce:
        msg = "Error setting up pssm screen"
        _LOGGER.error(f"{msg}: {exce}")
        _LOGGER.debug(msg = f"{type(exce)} info:", exc_info=exce)
        window.set_inkboard_state("ERROR")
        window.set_progress_bar(value=ttk.DANGER, text=msg)
    finally:
        if window._inkBoard_lock.locked():
            window._inkBoard_lock.release()
        if reload_finally: 
            unload_inkBoard(True)
    return

async def run_inkboard_config(configuration, **kwargs):
    if window._inkBoard_thread:
        stop_emulator(exce=QuitInkboard("Loading new config"),new_state=ttk.DISABLED)
        await asyncio.sleep(0)
        await asyncio.to_thread(window._inkBoard_thread.join)
        await asyncio.sleep(0)
        await asyncio.to_thread(window._inkBoard_lock.acquire)
        await asyncio.sleep(0)
        await asyncio.to_thread(window._inkBoard_lock.release)
        if not window._inkBoard_clean:
            await unload_inkBoard()

    await asyncio.to_thread(window._inkBoard_lock.acquire)
    window._current_config_file = configuration
    loop = asyncio.new_event_loop()
    thread = threading.Thread(target=loop.run_until_complete, kwargs={"future": run_inkboard_thread(configuration)}, name="inkBoard-thread")
    window._inkBoard_thread = thread
    thread.start()
    window._inkBoard_lock.release()
    return

async def async_run_designer(args):
    
    ##Settings for logging: same as run
    ##Config command: not required. For now, do not load anything if so, but will provide an option to open on the last opened config by default

    threading.excepthook = util.threading_except_hook
    tk_functions.runners = sys.modules[__name__]

    window = build_window()
    window.protocol("WM_DELETE_WINDOW", stop_designer)
    window.set_inkboard_state(None)

    if args.configuration:
        asyncio.create_task(run_inkboard_config(**vars(args)))
    await window.run_update_loop()
    return


"""
Provides base update functions (functions called when the entity associated with an element is updated) for some of the pssm elements.
"""
from typing import Callable, Union, TYPE_CHECKING, TypedDict, Optional, Any
import logging
import asyncio
from functools import partialmethod, partial
from inspect import getcoroutinestate, getcoroutinelocals
from ast import literal_eval

from PIL import Image

from PythonScreenStackManager import elements as elts, tools
import mdi_pil as mdi
from inkBoard import core as CORE

from .constants import DEFAULT_DOMAIN_ACTIONS, UNKNOWN_ICON, UNAVAILABLE_ICON, UNAVAILABLE_COLOR, UNKNOWN_COLOR, all_entities
from .helpers import triggerDictType, request_image_threadsafe

if TYPE_CHECKING:    
    from .HAelements import HAelement
    from .client import HAclient
    from PIL import Image
    import requests
    from PythonScreenStackManager import elements as elts
    from PythonScreenStackManager import tools

_LOGGER = logging.getLogger(__name__)

attribute_styles_stateType = TypedDict('attribute_style_states', {"state": Any, 'properties': dict})
attribute_stylesType = TypedDict('attribute_stylesDict', {'attribute': str, 'states': attribute_styles_stateType, 'else': dict})

state_color_dict = CORE.config.styles.get("state_colors",{})

def get_condition_key(state : str, conditions : list[str]):
    """
    Tests if any of the conditional strings in conditions return true. 
    Conditions can be defined as e.g. 'state < 5', the first one to evaluate as true is returned, otherwise None.
    state is the element state, i.e. if an attribute is set, it will set state to that attribute.
    """
    if isinstance(state,str):
        str_state = state
        try:
            ##First check if state can be converted into a different type
            state = literal_eval(str_state) #@IgnoreException
        except (SyntaxError, ValueError):
            ##state cannot be evaluated as something else than a string
            pass
    for cond in conditions:
        if "state" not in cond: continue
        try:
            if eval(cond): return cond #@IgnoreException
        except (SyntaxError, NameError, TypeError, ZeroDivisionError):
            continue
    
    ##Return default if no matches were found
    return "default" if "default" in conditions else None

def get_new_state(element : "HAelement", trigger_dict : "triggerDictType", skip_conditions : bool=False):
    """
    Returns the current state of the entity associated with this element.
    Mainly just present to prevent repetition in other trigger functions.
    args:
        element: the element to determine the state form
        trigger_dict: dict with the entity triggers
        skip_conditions (bool): If state_conditionals is true, setting this to true will mean the conditional test is not performed.
    """
    ##Handle eval states hiero
    ##First test if any of the expressions can be evaluated to true, else check for the state, else return default

    if attr := getattr(element,"entity_attribute", None):
        if attr in trigger_dict["to_state"]["attributes"]:
            new_state = trigger_dict["to_state"]["attributes"][attr]
            if not element.state_conditionals:
                new_state = str(new_state)
        else:
            if trigger_dict["to_state"]["state"] in ["unknown", "unavailable"]:
                new_state = trigger_dict["to_state"]["state"]
            else:
                new_state = None #"default"
    else:
        new_state = trigger_dict["to_state"]["state"]
    
    if element.state_conditionals and not skip_conditions: # and not new_state in element.state_styles:
        conditions = element.state_styles.keys()
        key = get_condition_key(new_state,conditions)
        new_state = key if key != None else new_state

    return new_state

def get_attribute_styles(element : "HAelement", trigger_dict : "triggerDictType") -> dict:
    """
    Function to check for each attribute in an elements attribute_styles if conditions matches.

    Parameters
    ----------
    element : HAelement
        The element to check
    trigger_dict : triggerDictType
        Entity's trigger

    Returns
    -------
    dict
        Dict with element properties to update
    """

    ##Maybe make a list for each attribute?
    ##i.e. - attribute: app
    ##       states:
    ##            - state: Netflix
    #               properties: blahblahblah
    ##       else:
    ##          color: blahblahblah
    attr_list = getattr(element, 'attribute_styles',[])
    if not attr_list: 
        return {}

    attr_states = trigger_dict["to_state"]["attributes"].copy()
    attr_states["state"] = trigger_dict["to_state"]["state"]

    upd_dict = {}
    for attr_conf in attr_list:
        prop_dict = {}
        attr = attr_conf.get("attribute", None)
        if attr == None:
            _LOGGER.warning(f"{element.id} attribute_styles is missing an attribute key at index {attr_list.index(attr_conf)}")
            continue

        # if attr not in attr_states or "states" not in attr_conf:
        if "states" not in attr_conf:
            if "else" in attr_conf:
                upd_dict.update(attr_conf["else"])
            continue

        state = attr_states.get(attr,None)
        if state == None:
            state = 'None'

        if attr == "state":
            str_state = state
            try:
                ##First check if state can be converted into a different type
                state = literal_eval(str_state) #@IgnoreException
            except (SyntaxError, ValueError):
                ##state cannot be evaluated as something else than a string
                pass

        for conf_state in attr_conf["states"]:
            if state == conf_state["state"]:
                prop_dict.update(conf_state.get("properties", {}))
                continue

            str_state = conf_state["state"]
            try:
                if eval(str_state): #@IgnoreException
                    prop_dict.update(conf_state.get("properties", {}))
            except (SyntaxError, NameError, TypeError, ZeroDivisionError):
                pass

        if not prop_dict and "else" in attr_conf:
            upd_dict.update(attr_conf["else"])
        elif prop_dict:
            upd_dict.update(prop_dict)
    
    return upd_dict

async def get_entity_picture(entity_picture : str, client : "HAclient") -> tuple[Union["Image.Image", "requests.Response"],int]:
    """
    Gets the entity_picture from an entity's entity_picture attribute.
    Appends the client's url if needed.   

    Parameters
    ----------
    entity_picture : str
        The entity_picture attribute, as gotten from a trigger_dict
    client : HAclient
        The Home Assistant client to get the picture from

    Returns
    -------
    Returns
    -------
    tuple[Image.Image | requests.response, status_code] | 
        If the status code is 200 (i.e. the request was succesfull) a tuple is returned with the gotten Image and the status code. 
        Otherwise a tuple with the full response and the status code is returned.
    """

    if "http" not in entity_picture:
        hass_url = client.hass_data["url"]
        ##I assume this should take care of getting a correct url?
        ##The SSL protocol should be present anyways if using an external url
        if not "http" in hass_url:
            hass_url = "http://" + hass_url

        url = hass_url + entity_picture
    else:
        url = entity_picture

    return await request_image_threadsafe(url)

async def button_trigger(element: Union["elts.Button", "HAelement"],trigger_dict : "triggerDictType"):
    """
    Default trigger function for button (text) elements
    
    Parameters
    ----------
    element : Union[&quot;elts.Button&quot;, &quot;HAelement&quot;]
        the element this trigger originated from
    trigger_dict : triggerDictType
        the dict with trigger information
    """
    ##For buttons, element states are not required, if not it simply sets the state as new text
    
    newAttributes = {}
    new_state = get_new_state(element,trigger_dict)
    if new_state not in element.state_styles:
        if new_state == "unknown":
            newAttributes["font_color"] = UNKNOWN_COLOR
        elif new_state == "unavailable":
            newAttributes["font_color"] = UNAVAILABLE_COLOR
        
        elif "else" in element.state_styles:
            new_state = "else"
    
    statedict = element.state_styles.get(new_state,False)

    if statedict:
        if isinstance(statedict,str):
            newAttributes["text"] = statedict
        else:
            newAttributes.update(statedict)
    else:
        newAttributes["text"] = new_state
        
    if new_state in state_color_dict and getattr(element,"state_colors",False):
        newAttributes.setdefault("font_color", state_color_dict[new_state])

    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)

    if newAttributes:
        skipPrint = getattr(element.parentPSSMScreen,"isBatch",False)
        await element.async_update(newAttributes, skipPrint=skipPrint)
    return

async def icon_trigger(element: Union["elts.Icon", "HAelement"],trigger_dict : "triggerDictType"):
    """
    Default trigger function for icon elements
    
    Parameters
    ----------
    element : Union[&quot;elts.Icon&quot;, &quot;HAelement&quot;]
        the element this trigger originated from
    trigger_dict : triggerDictType
        the dict with trigger information
    """

    newAttributes = {}
    new_state = get_new_state(element,trigger_dict)

    if new_state not in element.state_styles:
        if new_state == "unknown" or new_state == None:
            newAttributes["badge_icon"] = UNKNOWN_ICON
        elif new_state == "unavailable":
            newAttributes["badge_icon"] = UNAVAILABLE_ICON
        if new_state == None:
            new_state = "else"

    statedict = element.state_styles.get(new_state,False)

    if statedict:
        if isinstance(statedict,str):
            newAttributes["icon"] = statedict
        else:
            newAttributes.update(statedict)

    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)
        ##Check if this state is defined in the default state colors. If so, and state_colors is true and icon_color has not been set yet, set it.

    if "icon_attribute" in newAttributes: 
        icon_attr = newAttributes["icon_attribute"] 
    else:
        icon_attr = getattr(element, "icon_attribute", None)
    
    if icon_attr != None and "icon" not in newAttributes:
        ##Defining an icon in a state overwrites the attribute.
        _LOGGER.debug(f"{element}: Getting entity {element.entity} picture")
        if icon_attr == "entity_picture":
            if element.iconData[0] != trigger_dict["to_state"]["attributes"]["entity_picture"] or element.iconData[1] != 200:

                picture_link =  trigger_dict["to_state"]["attributes"]["entity_picture"]
                (resp, status) = await get_entity_picture(picture_link, client=element.HAclient)

                if status == 200:
                    newAttributes.update({'icon': resp})
                    element._iconData = (picture_link, status)
                else:
                    element._iconData = (picture_link, status)

        else:
            new_icon = trigger_dict["to_state"]["attributes"].get(icon_attr,None)
            if new_icon != None and element.icon != new_icon:
                newAttributes["icon"] = new_icon

    if new_state in state_color_dict and getattr(element,"state_colors",False):
        newAttributes.setdefault("icon_color", state_color_dict[new_state])

    if newAttributes:
        await element.async_update(newAttributes)
        if element.fileError:
            _LOGGER.warning(f"Icon {element.icon} for state {new_state} could not be found.")
    return

async def picture_trigger(element: Union["elts.Picture", "HAelement"],trigger_dict : "triggerDictType"):
    
    newAttributes = {}
    new_state = get_new_state(element,trigger_dict)
    ##Also allow for default picture; so add picture setter?
    ##Nope can be done in default state maybe

    if new_state not in element.state_styles:
        pass

    statedict = element.state_styles.get(new_state,False)
    if statedict:
        if isinstance(statedict,str):
            newAttributes["picture"] = statedict
        else:
            newAttributes.update(statedict)

    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)

    if "picture_attribute" in newAttributes: 
        pic_attr = newAttributes["picture_attribute"] 
    else:
        pic_attr = getattr(element, "picture_attribute", None)

    status = 0
    if pic_attr != None and "picture" not in newAttributes:
        ##Defining an icon in a state overwrites the attribute.
        _LOGGER.verbose(f"{element}: Getting picture")

        ##For pictures, I think it makes more sense to assume you're getting an image from the HA server.
        ##So we always try and get it.
        ##Don't know if the link works like this for other entities/attributes but eh we'll see.
        if pic_attr not in trigger_dict["to_state"]["attributes"]:
            icon = trigger_dict["to_state"]["attributes"].get("icon", None)
            if icon == None:
                if hasattr(element,"fallback_icon"):
                    icon = element.fallback_icon
                else:
                    icon = elts.MISSING_PICTURE_ICON

            if element.area != None:
                [_, size] = element.area
            else:
                size = (100,100)
            
            if mdi.is_mdi(icon):
                img = Image.new("RGBA", size, element.background_color)
                img = mdi.draw_mdi_icon(img, icon, icon_size=int(size[1]*0.3))
            else:
                img = icon

            newAttributes.update({'picture': img})
            element.pictureData = (None, -1)
        elif (element.pictureData == None or
            element.pictureData[0] != trigger_dict["to_state"]["attributes"][pic_attr] or 
            element.pictureData[1] != 200):

            picture_link =  trigger_dict["to_state"]["attributes"][pic_attr]
            (resp, status) = await get_entity_picture(picture_link, client=element.HAclient)

            if status == 200:
                newAttributes.update({'picture': resp})
                element.pictureData = (picture_link, status)
            else:
                element.pictureData = (picture_link, status)

    if newAttributes:
        skipPrint = getattr(element.parentPSSMScreen,"isBatch",False)
        await element.async_update(newAttributes, skipPrint=skipPrint)
        if element.fileError:
            _LOGGER.warning(f"Picture {element.picture} for state {new_state} could not be found.")
    if status == 200:
        return

async def slider_trigger(element: Union["elts.Slider", "HAelement"],trigger_dict : "triggerDictType"):
    """
    Default trigger functions for sliders
    
    Parameters
    ----------
    element : Union[&quot;elts.Counter&quot;, &quot;HAelement&quot;]
        the element this trigger originated from
    trigger_dict : triggerDictType
        the dict with trigger information
    """
    ##Use min_attribute and max_attribute for the min/max values
    if element.minAttribute != None:
        if isinstance(element.minAttribute,str):
            minVal = trigger_dict["to_state"]["attributes"].get(element.minAttribute,None)
        else:
            minVal = element.minAttribute
        if minVal != None and element.minimum != minVal:
            element.minimum = minVal

    if element.maxAttribute != None:
        if isinstance(element.maxAttribute,str):
            maxVal = trigger_dict["to_state"]["attributes"].get(element.maxAttribute,None)
        else:
            maxVal = element.maxAttribute
        if maxVal != None and element.maximum != maxVal:
            element.maximum = maxVal
    new_state = get_new_state(element,trigger_dict)

    newAttributes = {}

    if new_state  in ["off", "unknown", "unavailable", None, 'None']:
        
        ##What to do here for unavailable etc?
        position = element.minimum
        if new_state in ["unavailable", "unknown"]:
            icon = UNAVAILABLE_ICON if new_state == "unavailable" else UNKNOWN_ICON
            if isinstance(e := element.endPoints,str):
                ep = e
            else:
                ep = e[1]
            newAttributes["endPoints"] = (icon, ep)
            
    else:
        try:
            position = float(new_state) #@IgnoreException
            ##Rewrite sliders to allow setting position within range.
        except TypeError:
            position = element.minimum
            _LOGGER.exception(f"New state {new_state} for slider {element.id} could not be converted into a number.")

    newAttributes.update({'position': position})

    statedict = element.state_styles.get(new_state,False)

    if statedict:
        newAttributes.update(statedict)    

    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)

    if newAttributes:
        skipPrint = getattr(element.parentPSSMScreen,"isBatch",False)
        await element.async_update(newAttributes, skipPrint=skipPrint)

async def counter_trigger(element: Union["elts.Counter", "HAelement"],trigger_dict : "triggerDictType"):
    """
    Default trigger function for counter elements

    Parameters
    ----------
    element : Union[&quot;elts.Counter&quot;, &quot;HAelement&quot;]
        the element this trigger originated from
    trigger_dict : triggerDictType
        the dict with trigger information
    """

    newAttributes = {}
    
    if element.minAttribute != None:
        if isinstance(element.minAttribute,str):
            minVal = trigger_dict["to_state"]["attributes"].get(element.minAttribute,None)
        else:
            minVal = element.minAttribute
        if minVal != None and element.minimum != minVal:
            element.minimum = minVal

    if element.maxAttribute != None:
        if isinstance(element.maxAttribute,str):
            maxVal = trigger_dict["to_state"]["attributes"].get(element.maxAttribute,None)
        else:
            maxVal = element.maxAttribute
        if maxVal != None and element.maximum != maxVal:
            element.maximum = maxVal

    if element.stepAttribute != None:
        if isinstance(element.stepAttribute,str):
            stepVal = trigger_dict["to_state"]["attributes"].get(element.stepAttribute,None)
        else:
            stepVal = element.stepAttribute
        if stepVal != None and element.step != stepVal:
            element.step = stepVal

    new_state = get_new_state(element, trigger_dict)
    newAttributes["value"] = float(new_state)
    
    statedict = element.state_styles.get(new_state,False)
    if statedict:
        newAttributes.update(statedict)
    
    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)

    if newAttributes:
        skipPrint = getattr(element.parentPSSMScreen,"isBatch",False)
        await element.async_update(newAttributes, skipPrint=skipPrint)

    return

async def select_trigger(element: Union["elts.DropDown", "HAelement"], trigger_dict : "triggerDictType"):
    """
    Default trigger function for (menu)select elements

    Parameters
    ----------
    element : Union[&quot;elts.DropDown&quot;, &quot;HAelement&quot;]
        the element this trigger originated from
    trigger_dict : triggerDictType
        the dict with trigger information
    """
    
    if element.optionsAttribute != None:
        if isinstance(element.optionsAttribute,str):
            options = trigger_dict["to_state"]["attributes"].get(element.optionsAttribute,None)
        else:
            options = element.optionsAttribute
        if options != None and element.options != options:
            element.options = options

    newAttributes = {}
    new_state = get_new_state(element,trigger_dict)
    
    if new_state  in ["unknown", "unavailable"]:

        icon = UNAVAILABLE_ICON if new_state == "unavailable" else UNKNOWN_ICON
        newAttributes["closedIcon"] = icon
    
    if new_state in element.options:
        idx = element.options.index(new_state)
        newAttributes["_selected"] = idx

    statedict = element.state_styles.get(new_state,False)
    if statedict:
        newAttributes.update(statedict)

    if hasattr(element,"attribute_styles"):
        attr_props = get_attribute_styles(element, trigger_dict)
        newAttributes.update(attr_props)

    if newAttributes:
        skipPrint = getattr(element.parentPSSMScreen,"isBatch",False)
        await element.async_update(newAttributes, skipPrint=skipPrint)

    
    return

element_triggers = {
    elts.Icon: icon_trigger,
    elts.Picture: picture_trigger,
    elts.Button: button_trigger,
    elts.Counter: counter_trigger,
    elts.Slider: slider_trigger,
    elts.BoxSlider: slider_trigger,
    elts.LineSlider: slider_trigger,
    elts.DropDown: select_trigger,
    ##Should perhaps include booleans in here too? For switches e.g.
}
"Links baseelements to default trigger functions"
#endregion

#region Setter functions
tosetCounter = TypedDict("toset",{"entity_attribute" : Optional[str], "minAttribute" : Union[str,int,float] ,"maxAttribute" : Union[str,int,float], "stepAttribute" : Union[str,int,float], 
                            "action" : str, "action_data_map": dict[str,str]}, total=False)
tosetCounter.__required_keys__ = frozenset({'entity_attribute','minAttribute','maxAttribute'})
def min_max_setter(element : Union["HAelement",elts.Slider, elts.BoxSlider, elts.Counter], savedattributes : dict = {}):
    """
    Default setter for Counter and Slider elements with entities attached

    Parameters
    ----------
    element : Union[elts.Slider, elts.BoxSlider, elts.Counter]
        the element that will be set
    savedattributes : dict, optional
        attributes that were saved during the wrapping of this element into a home assistant element, by default {}
    """
    
    toset = tosetCounter
    defaultDomains = {"counter" :  toset(entity_attribute=None,minAttribute="minimum",maxAttribute="maximum", stepAttribute="step", action="counter.set_value", action_data_map={"value":"value"}), 
                    "number":  toset(entity_attribute=None,minAttribute="min",maxAttribute="max", stepAttribute="step", action="number.set_value", action_data_map={"value":"value"}),
                    "input_number":  toset(entity_attribute=None,minAttribute="min",maxAttribute="max", stepAttribute="step", action="input_number.set_value", action_data_map={"value":"value"}),
                    "light":  toset(entity_attribute="brightness",minAttribute=0,maxAttribute=254, stepAttribute=1, action="light.turn_on", action_data_map={"brightness": "value"}), ##no min/max brightness so how to set that -> I think just set an optional number value
                    "climate":  toset(entity_attribute="temperature",minAttribute="min_temp",maxAttribute="max_temp", stepAttribute="target_temp_step", action="climate.set_temperature", action_data_map={"temperature": "value"}),
                    "fan":  toset(entity_attribute="percentage",minAttribute=0,maxAttribute=100, stepAttribute=1, action="fan.set_percentage", action_data_map={"percentage": "value"}),
                    "cover" :  toset(entity_attribute="current_cover_position",minAttribute=0,maxAttribute=100, stepAttribute=1, action="cover.set_cover_position", action_data_map={"position": "value"}),
                    "media_player" :  toset(entity_attribute="volume_level",minAttribute=0,maxAttribute=1, stepAttribute=0.1, action="media_player.volume_set", action_data_map={"volume_level": "value"}),
                    "water_heater" :  toset(entity_attribute="target_temperature",minAttribute="min_temp",maxAttribute="max_temp", action="water_heater.set_temperature", action_data_map={"temperature": "value"}),
                    "valve":  toset(entity_attribute="current_valve_position",minAttribute=0,maxAttribute=100,step=1,action="valve.set_position", action_data_map={"position":"value"})
                    }
    ##Most unknown domains use percentages. media_player is here for volume support
    domain = element.entity.split(".")[0]
    if domain not in defaultDomains:
        return
    
    if isinstance(element, elts.Counter):
        toset.__required_keys__ = toset.__required_keys__.union(frozenset({"stepAttribute"}))
        if element.on_count == None:
            # element.on_count = element.HAclient.call_service_action
            v = defaultDomains[domain]["action"]
            d = {"action": "service-action", 
                "data": {"action": defaultDomains[domain]["action"]}, "map": defaultDomains[domain]["action_data_map"]}
            element.on_count = d
            ##So set on_count_map and on_count_data
            ##service_action_data should be made from the kwargs
    elif isinstance(element, (elts.Slider, elts.BoxSlider)):
        if element._tap_action == None:
            # element.tap_action = element.HAclient.call_service_action
            d = {"action": "service-action", 
                "data": {"action": defaultDomains[domain]["action"]}, "map": defaultDomains[domain]["action_data_map"]}
            element.tap_action = d

    for attr in toset.__required_keys__:
        if attr in savedattributes:
            ##Declare an attribute when building the element as None to not set a default.
            continue
        
        v = defaultDomains[domain][attr]
        setattr(element,attr,v)

    return

def options_setter(element : Union["HAelement", elts.DropDown], savedattributes : dict = {}):
    toset = TypedDict("toset",{"entity_attribute" : Optional[str], "optionsAttribute" : Union[str,int,float] ,
                            "action" : str, "action_data_map": dict[str,str]}, total=False)
    toset.__required_keys__ = frozenset({'entity_attribute','optionsAttribute'})
    defaultDomains = {"select" :  toset(entity_attribute=None,optionsAttribute="options", action="select.select_option", action_data_map={"option":"selected_option"}), 
                    "input_select":  toset(entity_attribute=None,optionsAttribute="options", action="input_select.select_option", action_data_map={"option":"selected_option"})
    }

    domain = element.entity.split(".")[0]
    if domain not in defaultDomains:
        return

    if element.on_select == None:
        element.on_select = {
            "action": element.HAclient.call_service_action,
            "data": {"action": defaultDomains[domain]["action"]},
            "map": defaultDomains[domain]["action_data_map"]
        }
        toset.__required_keys__ = toset.__required_keys__.union(frozenset({"action", "action_data_map"}))

    if "action" in savedattributes and "action_data_map" not in savedattributes: savedattributes["action_data_map"] = {}
    for attr in toset.__required_keys__:
        if attr in savedattributes:
            ##Declare an attribute when building the element as None to not set a default.
            continue
        
        v = defaultDomains[domain][attr]
        setattr(element,attr,v)

    return

def default_action_setter(element : "HAelement"):
    """
    Sets the default tap_action for some base elements according to their entity.

    Parameters
    ----------
    element : HAelement
        the element to set the tap_action of.
    """

    if not isinstance(element.entity,str) or getattr(element, "HAclient", None) == None:
        return

    domain = element.entity.split(".")[0]
    if domain in DEFAULT_DOMAIN_ACTIONS and element.tap_action == None:
        d = {"action": "service-action"}

        if domain in ["input_select", "select"]:
            d["data"] = {"action": DEFAULT_DOMAIN_ACTIONS[domain], "cycle": True}
        else:
            d["data"] = {"action": DEFAULT_DOMAIN_ACTIONS[domain]}
        
        element.tap_action = d

compound_setters = {
    elts.Counter: min_max_setter,
    elts.Slider: min_max_setter,
    elts.BoxSlider: min_max_setter,
    elts.DropDown: options_setter,
}
##Should I include a setter here for the bool select?

default_tap_action_setters = {
    elts.Button: default_action_setter, 
    elts.Icon: default_action_setter,
    elts.Picture: default_action_setter
    }

def set_trigger_function(element : "HAelement", savedattributes : dict = {}) -> Callable:
    """
    Applies the appropriate default trigger function to an element if none have been set.

    Parameters
    ----------
    element : HAelement
        element to apply the function to
    savedattributes : dict, optional
        attributes that were already applied to the element before wrapping, by default {}

    Returns
    -------
    Callable
        _description_
    """
    
    "Sets the base trigger function for the provided element type if it was not yet set."
    eCls =  element.__class__.__base__
    if "trigger_function" in savedattributes:
        ##Perform the check here for possible custom functions
        if isinstance(element.trigger_function,str):
            if "custom:" in element.trigger_function.lower():
                ##Will not do this here, but do it via the parser.
                pass
        ##Probably not return here yet
        # return
    elif eCls in element_triggers:
        element.trigger_function = element_triggers[eCls] #element_triggers[element.__class__.__base__]
    else:
        _LOGGER.info(f"Got pssm element without base trigger function: {element.__class__.__name__}")
    
    if eCls in compound_setters:
        setter = compound_setters[eCls]
        setter(element, savedattributes)

    if eCls in default_tap_action_setters and element.tap_action == None:
        func = default_tap_action_setters[eCls]
        func(element)

    return None 

#endregion